"""Code regarding the PEtab YAML config files"""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import jsonschema
import numpy as np
import yaml
from pandas.io.common import get_handle

from .C import *  # noqa: F403

# directory with PEtab yaml schema files
SCHEMA_DIR = Path(__file__).parent / "schemas"
# map of version number to validation schema
SCHEMAS = {
    "1": SCHEMA_DIR / "petab_schema.v1.0.0.yaml",
    "1.0.0": SCHEMA_DIR / "petab_schema.v1.0.0.yaml",
    "2.0.0": SCHEMA_DIR / "petab_schema.v2.0.0.yaml",
}

__all__ = [
    "validate",
    "validate_yaml_syntax",
    "validate_yaml_semantics",
    "load_yaml",
    "is_composite_problem",
    "assert_single_condition_and_sbml_file",
    "write_yaml",
    "create_problem_yaml",
]


def validate(
    yaml_config: Union[Dict, str, Path],
    path_prefix: Union[None, str, Path] = None,
):
    """Validate syntax and semantics of PEtab config YAML

    Arguments:
        yaml_config:
            PEtab YAML config as filename or dict.
        path_prefix:
            Base location for relative paths. Defaults to location of YAML
            file if a filename was provided for ``yaml_config`` or the current
            working directory.
    """

    validate_yaml_syntax(yaml_config)
    validate_yaml_semantics(yaml_config=yaml_config, path_prefix=path_prefix)


def validate_yaml_syntax(
    yaml_config: Union[Dict, str, Path], schema: Union[None, Dict, str] = None
):
    """Validate PEtab YAML file syntax

    Arguments:
        yaml_config:
            PEtab YAML file to validate, as file name or dictionary
        schema:
            Custom schema for validation

    Raises:
        see :func:`jsonschema.validate`
    """
    yaml_config = load_yaml(yaml_config)

    if schema is None:
        # try get PEtab version from yaml file
        #  if this is not the available, the file is not valid anyways,
        #  but let's still use the latest PEtab schema for full validation
        version = (
            yaml_config.get(FORMAT_VERSION, None) or list(SCHEMAS.values())[-1]
        )
        try:
            schema = SCHEMAS[str(version)]
        except KeyError as e:
            raise ValueError(
                "Unknown PEtab version given in problem "
                f"specification: {version}"
            ) from e
    schema = load_yaml(schema)
    jsonschema.validate(instance=yaml_config, schema=schema)


def validate_yaml_semantics(
    yaml_config: Union[Dict, str, Path],
    path_prefix: Union[None, str, Path] = None,
):
    """Validate PEtab YAML file semantics

    Check for existence of files. Assumes valid syntax.

    Version number and contents of referenced files are not yet checked.

    Arguments:
        yaml_config:
            PEtab YAML config as filename or dict.
        path_prefix:
            Base location for relative paths. Defaults to location of YAML
            file if a filename was provided for ``yaml_config`` or the current
            working directory.

    Raises:
        AssertionError: in case of problems
    """
    if not path_prefix:
        if isinstance(yaml_config, (str, Path)):
            path_prefix = os.path.dirname(str(yaml_config))
        else:
            path_prefix = ""

    yaml_config = load_yaml(yaml_config)

    def _check_file(_filename: str, _field: str):
        if not os.path.isfile(_filename):
            raise AssertionError(
                f"File '{_filename}' provided as '{_field}' " "does not exist."
            )

    # Handles both a single parameter file, and a parameter file that has been
    # split into multiple subset files.
    for parameter_subset_file in list(
        np.array(yaml_config[PARAMETER_FILE]).flat
    ):
        _check_file(
            os.path.join(path_prefix, parameter_subset_file),
            parameter_subset_file,
        )

    for problem_config in yaml_config[PROBLEMS]:
        for field in [
            SBML_FILES,
            CONDITION_FILES,
            MEASUREMENT_FILES,
            VISUALIZATION_FILES,
            OBSERVABLE_FILES,
        ]:
            if field in problem_config:
                for filename in problem_config[field]:
                    _check_file(os.path.join(path_prefix, filename), field)


def load_yaml(yaml_config: Union[Dict, Path, str]) -> Dict:
    """Load YAML

    Convenience function to allow for providing YAML inputs as filename, URL
    or as dictionary.

    Arguments:
        yaml_config:
            PEtab YAML config as filename or dict or URL.

    Returns:
        The unmodified dictionary if ``yaml_config`` was dictionary.
        Otherwise the parsed the YAML file.
    """

    # already parsed? all PEtab problem yaml files are dictionaries
    if isinstance(yaml_config, dict):
        return yaml_config

    with get_handle(yaml_config, mode="r") as io_handle:
        data = yaml.safe_load(io_handle.handle)
    return data


def is_composite_problem(yaml_config: Union[Dict, str, Path]) -> bool:
    """Does this YAML file comprise multiple models?

    Arguments:
        yaml_config: PEtab configuration as dictionary or YAML file name
    """

    yaml_config = load_yaml(yaml_config)
    return len(yaml_config[PROBLEMS]) > 1


def assert_single_condition_and_sbml_file(problem_config: Dict) -> None:
    """Check that there is only a single condition file and a single SBML
    file specified.

    Arguments:
        problem_config:
            Dictionary as defined in the YAML schema inside the `problems`
            list.
    Raises:
        NotImplementedError:
            If multiple condition or SBML files specified.
    """
    if (
        len(problem_config[SBML_FILES]) > 1
        or len(problem_config[CONDITION_FILES]) > 1
    ):
        # TODO https://github.com/ICB-DCM/PEtab/issues/188
        # TODO https://github.com/ICB-DCM/PEtab/issues/189
        raise NotImplementedError(
            "Support for multiple models or condition files is not yet "
            "implemented."
        )


def write_yaml(
    yaml_config: Dict[str, Any], filename: Union[str, Path]
) -> None:
    """Write PEtab YAML file

    Arguments:
        yaml_config: Data to write
        filename: File to create
    """
    with open(filename, "w") as outfile:
        yaml.dump(
            yaml_config, outfile, default_flow_style=False, sort_keys=False
        )


def create_problem_yaml(
    sbml_files: Union[str, Path, List[Union[str, Path]]],
    condition_files: Union[str, Path, List[Union[str, Path]]],
    measurement_files: Union[str, Path, List[Union[str, Path]]],
    parameter_file: Union[str, Path],
    observable_files: Union[str, Path, List[Union[str, Path]]],
    yaml_file: Union[str, Path],
    visualization_files: Optional[
        Union[str, Path, List[Union[str, Path]]]
    ] = None,
    relative_paths: bool = True,
    mapping_files: Union[str, Path, List[Union[str, Path]]] = None,
) -> None:
    """Create and write default YAML file for a single PEtab problem

    Arguments:
        sbml_files: Path of SBML model file or list of such
        condition_files: Path of condition file or list of such
        measurement_files: Path of measurement file or list of such
        parameter_file: Path of parameter file
        observable_files: Path of observable file or list of such
        yaml_file: Path to which YAML file should be written
        visualization_files:
            Optional Path to visualization file or list of such
        relative_paths:
            whether all paths in the YAML file should be relative to the
            location of the YAML file. If ``False``, then paths are left
            unchanged.
        mapping_files: Path of mapping file
    """
    if isinstance(sbml_files, (Path, str)):
        sbml_files = [sbml_files]
    if isinstance(condition_files, (Path, str)):
        condition_files = [condition_files]
    if isinstance(measurement_files, (Path, str)):
        measurement_files = [measurement_files]
    if isinstance(observable_files, (Path, str)):
        observable_files = [observable_files]
    if isinstance(visualization_files, (Path, str)):
        visualization_files = [visualization_files]

    if relative_paths:
        yaml_file_dir = Path(yaml_file).parent

        def get_rel_to_yaml(paths: Union[List[str], None]):
            if paths is None:
                return paths
            return [
                os.path.relpath(path, start=yaml_file_dir) for path in paths
            ]

        sbml_files = get_rel_to_yaml(sbml_files)
        condition_files = get_rel_to_yaml(condition_files)
        measurement_files = get_rel_to_yaml(measurement_files)
        observable_files = get_rel_to_yaml(observable_files)
        visualization_files = get_rel_to_yaml(visualization_files)
        parameter_file = get_rel_to_yaml([parameter_file])[0]
        mapping_files = get_rel_to_yaml(mapping_files)

    problem_dic = {
        CONDITION_FILES: condition_files,
        MEASUREMENT_FILES: measurement_files,
        SBML_FILES: sbml_files,
        OBSERVABLE_FILES: observable_files,
    }
    if mapping_files:
        problem_dic[MAPPING_FILES] = mapping_files

    if visualization_files is not None:
        problem_dic[VISUALIZATION_FILES] = visualization_files
    yaml_dic = {
        PARAMETER_FILE: parameter_file,
        FORMAT_VERSION: 1,
        PROBLEMS: [problem_dic],
    }
    write_yaml(yaml_dic, yaml_file)



"""PEtab problems consisting of multiple models"""
import os
from typing import Dict, List, Union

import pandas as pd

from . import parameters, problem, yaml
from .C import *  # noqa: F403

__all__ = ["CompositeProblem"]


class CompositeProblem:
    """Representation of a PEtab problem consisting of multiple models

    Attributes:
        problems:
            List of :py:class:`petab.Problem` s
        parameter_df:
            PEtab parameter DataFrame
    """

    def __init__(
        self,
        parameter_df: pd.DataFrame = None,
        problems: List[problem.Problem] = None,
    ):
        """Constructor

        Arguments:
            parameter_df:
                see CompositeProblem.parameter_df
            problems:
                see CompositeProblem.problems
        """
        self.problems: List[problem.Problem] = problems
        self.parameter_df: pd.DataFrame = parameter_df

    @staticmethod
    def from_yaml(yaml_config: Union[Dict, str]) -> "CompositeProblem":
        """Create from YAML file

        Factory method to create a CompositeProblem instance from a PEtab
        YAML config file

        Arguments:
            yaml_config: PEtab configuration as dictionary or YAML file name
        """

        if isinstance(yaml_config, str):
            path_prefix = os.path.dirname(yaml_config)
            yaml_config = yaml.load_yaml(yaml_config)
        else:
            path_prefix = ""

        parameter_df = parameters.get_parameter_df(
            os.path.join(path_prefix, yaml_config[PARAMETER_FILE])
        )

        problems = []
        for problem_config in yaml_config[PROBLEMS]:
            yaml.assert_single_condition_and_sbml_file(problem_config)

            # don't set parameter file if we have multiple models
            cur_problem = problem.Problem.from_files(
                sbml_file=os.path.join(
                    path_prefix, problem_config[SBML_FILES][0]
                ),
                measurement_file=[
                    os.path.join(path_prefix, f)
                    for f in problem_config[MEASUREMENT_FILES]
                ],
                condition_file=os.path.join(
                    path_prefix, problem_config[CONDITION_FILES][0]
                ),
                visualization_files=[
                    os.path.join(path_prefix, f)
                    for f in problem_config[VISUALIZATION_FILES]
                ],
                observable_files=[
                    os.path.join(path_prefix, f)
                    for f in problem_config[OBSERVABLE_FILES]
                ],
            )
            problems.append(cur_problem)

        return CompositeProblem(parameter_df=parameter_df, problems=problems)



from pathlib import Path
from tempfile import TemporaryDirectory

import petab
from petab.visualize.data_overview import create_report


def test_data_overview():
    """Data overview generation with Fujita example data from this repository"""
    with TemporaryDirectory() as temp_dir:
        outfile = Path(temp_dir) / "Fujita.html"
        repo_root = Path(__file__).parent.parent
        yaml_filename = (
            repo_root / "doc" / "example" / "example_Fujita" / "Fujita.yaml"
        )
        problem = petab.Problem.from_yaml(yaml_filename)
        create_report(problem, "Fujita", output_path=temp_dir)
        assert outfile.is_file()



"""PEtab Problem class"""
from __future__ import annotations

import os
import tempfile
from math import nan
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union
from urllib.parse import unquote, urlparse, urlunparse
from warnings import warn

import pandas as pd

from . import (
    conditions,
    core,
    format_version,
    mapping,
    measurements,
    observables,
    parameter_mapping,
    parameters,
    sampling,
    sbml,
    yaml,
)
from .C import *  # noqa: F403
from .models import MODEL_TYPE_SBML
from .models.model import Model, model_factory
from .models.sbml_model import SbmlModel

if TYPE_CHECKING:
    import libsbml


__all__ = ["Problem"]


class Problem:
    """
    PEtab parameter estimation problem as defined by

    - model
    - condition table
    - measurement table
    - parameter table
    - observables table
    - mapping table

    Optionally it may contain visualization tables.

    Attributes:
        condition_df: PEtab condition table
        measurement_df: PEtab measurement table
        parameter_df: PEtab parameter table
        observable_df: PEtab observable table
        visualization_df: PEtab visualization table
        mapping_df: PEtab mapping table
        model: The underlying model
        sbml_reader: Stored to keep object alive (deprecated).
        sbml_document: Stored to keep object alive (deprecated).
        sbml_model: PEtab SBML model (deprecated)
        extensions_config: Information on the extensions used
    """

    def __init__(
        self,
        sbml_model: libsbml.Model = None,
        sbml_reader: libsbml.SBMLReader = None,
        sbml_document: libsbml.SBMLDocument = None,
        model: Model = None,
        model_id: str = None,
        condition_df: pd.DataFrame = None,
        measurement_df: pd.DataFrame = None,
        parameter_df: pd.DataFrame = None,
        visualization_df: pd.DataFrame = None,
        observable_df: pd.DataFrame = None,
        mapping_df: pd.DataFrame = None,
        extensions_config: Dict = None,
    ):
        self.condition_df: Optional[pd.DataFrame] = condition_df
        self.measurement_df: Optional[pd.DataFrame] = measurement_df
        self.parameter_df: Optional[pd.DataFrame] = parameter_df
        self.visualization_df: Optional[pd.DataFrame] = visualization_df
        self.observable_df: Optional[pd.DataFrame] = observable_df
        self.mapping_df: Optional[pd.DataFrame] = mapping_df

        if any(
            (sbml_model, sbml_document, sbml_reader),
        ):
            warn(
                "Passing `sbml_model`, `sbml_document`, or `sbml_reader` "
                "to petab.Problem is deprecated and will be removed in a "
                "future version. Use `model=petab.models.sbml_model."
                "SbmlModel(...)` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if model:
                raise ValueError(
                    "Must only provide one of (`sbml_model`, "
                    "`sbml_document`, `sbml_reader`) or `model`."
                )

            model = SbmlModel(
                sbml_model=sbml_model,
                sbml_reader=sbml_reader,
                sbml_document=sbml_document,
                model_id=model_id,
            )

        self.model: Optional[Model] = model
        self.extensions_config = extensions_config or {}

    def __getattr__(self, name):
        # For backward-compatibility, allow access to SBML model related
        #  attributes now stored in self.model
        if name in {"sbml_model", "sbml_reader", "sbml_document"}:
            return getattr(self.model, name) if self.model else None
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no " f"attribute '{name}'"
        )

    def __setattr__(self, name, value):
        # For backward-compatibility, allow access to SBML model related
        #  attributes now stored in self.model
        if name in {"sbml_model", "sbml_reader", "sbml_document"}:
            if self.model:
                setattr(self.model, name, value)
            else:
                self.model = SbmlModel(**{name: value})
        else:
            super().__setattr__(name, value)

    def __str__(self):
        model = f"with model ({self.model})" if self.model else "without model"
        conditions = (
            f"{self.condition_df.shape[0]} conditions"
            if self.condition_df is not None
            else "without conditions table"
        )

        observables = (
            f"{self.observable_df.shape[0]} observables"
            if self.observable_df is not None
            else "without observables table"
        )

        measurements = (
            f"{self.measurement_df.shape[0]} measurements"
            if self.measurement_df is not None
            else "without measurements table"
        )

        if self.parameter_df is not None:
            num_estimated_parameters = (
                sum(self.parameter_df[ESTIMATE] == 1)
                if ESTIMATE in self.parameter_df
                else self.parameter_df.shape[0]
            )
            parameters = f"{num_estimated_parameters} estimated parameters"
        else:
            parameters = "without parameter_df table"

        return (
            f"PEtab Problem {model}, {conditions}, {observables}, "
            f"{measurements}, {parameters}"
        )

    @staticmethod
    def from_files(
        sbml_file: Union[str, Path] = None,
        condition_file: Union[str, Path, Iterable[Union[str, Path]]] = None,
        measurement_file: Union[str, Path, Iterable[Union[str, Path]]] = None,
        parameter_file: Union[str, Path, Iterable[Union[str, Path]]] = None,
        visualization_files: Union[
            str, Path, Iterable[Union[str, Path]]
        ] = None,
        observable_files: Union[str, Path, Iterable[Union[str, Path]]] = None,
        model_id: str = None,
        extensions_config: Dict = None,
    ) -> "Problem":
        """
        Factory method to load model and tables from files.

        Arguments:
            sbml_file: PEtab SBML model
            condition_file: PEtab condition table
            measurement_file: PEtab measurement table
            parameter_file: PEtab parameter table
            visualization_files: PEtab visualization tables
            observable_files: PEtab observables tables
            model_id: PEtab ID of the model
            extensions_config: Information on the extensions used
        """
        warn(
            "petab.Problem.from_files is deprecated and will be removed in a "
            "future version. Use `petab.Problem.from_yaml instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        model = (
            model_factory(sbml_file, MODEL_TYPE_SBML, model_id=model_id)
            if sbml_file
            else None
        )

        condition_df = (
            core.concat_tables(condition_file, conditions.get_condition_df)
            if condition_file
            else None
        )

        # If there are multiple tables, we will merge them
        measurement_df = (
            core.concat_tables(
                measurement_file, measurements.get_measurement_df
            )
            if measurement_file
            else None
        )

        parameter_df = (
            parameters.get_parameter_df(parameter_file)
            if parameter_file
            else None
        )

        # If there are multiple tables, we will merge them
        visualization_df = (
            core.concat_tables(visualization_files, core.get_visualization_df)
            if visualization_files
            else None
        )

        # If there are multiple tables, we will merge them
        observable_df = (
            core.concat_tables(observable_files, observables.get_observable_df)
            if observable_files
            else None
        )

        return Problem(
            model=model,
            condition_df=condition_df,
            measurement_df=measurement_df,
            parameter_df=parameter_df,
            observable_df=observable_df,
            visualization_df=visualization_df,
            extensions_config=extensions_config,
        )

    @staticmethod
    def from_yaml(yaml_config: Union[Dict, Path, str]) -> "Problem":
        """
        Factory method to load model and tables as specified by YAML file.

        Arguments:
            yaml_config: PEtab configuration as dictionary or YAML file name
        """
        if isinstance(yaml_config, Path):
            yaml_config = str(yaml_config)

        get_path = lambda filename: filename  # noqa: E731
        if isinstance(yaml_config, str):
            yaml_path = yaml_config
            yaml_config = yaml.load_yaml(yaml_config)

            # yaml_config may be path or URL
            path_url = urlparse(yaml_path)
            if not path_url.scheme or (
                path_url.scheme != "file" and not path_url.netloc
            ):
                # a regular file path string
                path_prefix = Path(yaml_path).parent
                get_path = (
                    lambda filename: path_prefix / filename
                )  # noqa: E731
            else:
                # a URL
                # extract parent path from
                url_path = unquote(urlparse(yaml_path).path)
                parent_path = str(PurePosixPath(url_path).parent)
                path_prefix = urlunparse(
                    (
                        path_url.scheme,
                        path_url.netloc,
                        parent_path,
                        path_url.params,
                        path_url.query,
                        path_url.fragment,
                    )
                )
                # need "/" on windows, not "\"
                get_path = (
                    lambda filename: f"{path_prefix}/{filename}"
                )  # noqa: E731

        if yaml.is_composite_problem(yaml_config):
            raise ValueError(
                "petab.Problem.from_yaml() can only be used for "
                "yaml files comprising a single model. "
                "Consider using "
                "petab.CompositeProblem.from_yaml() instead."
            )

        if yaml_config[FORMAT_VERSION] not in {"1", 1, "1.0.0", "2.0.0"}:
            raise ValueError(
                "Provided PEtab files are of unsupported version "
                f"{yaml_config[FORMAT_VERSION]}. Expected "
                f"{format_version.__format_version__}."
            )
        if yaml_config[FORMAT_VERSION] == "2.0.0":
            warn("Support for PEtab2.0 is experimental!")

        problem0 = yaml_config["problems"][0]

        if isinstance(yaml_config[PARAMETER_FILE], list):
            parameter_df = parameters.get_parameter_df(
                [get_path(f) for f in yaml_config[PARAMETER_FILE]]
            )
        else:
            parameter_df = (
                parameters.get_parameter_df(
                    get_path(yaml_config[PARAMETER_FILE])
                )
                if yaml_config[PARAMETER_FILE]
                else None
            )

        if yaml_config[FORMAT_VERSION] in [1, "1", "1.0.0"]:
            if len(problem0[SBML_FILES]) > 1:
                # TODO https://github.com/PEtab-dev/libpetab-python/issues/6
                raise NotImplementedError(
                    "Support for multiple models is not yet implemented."
                )

            model = (
                model_factory(
                    get_path(problem0[SBML_FILES][0]),
                    MODEL_TYPE_SBML,
                    model_id=None,
                )
                if problem0[SBML_FILES]
                else None
            )
        else:
            if len(problem0[MODEL_FILES]) > 1:
                # TODO https://github.com/PEtab-dev/libpetab-python/issues/6
                raise NotImplementedError(
                    "Support for multiple models is not yet implemented."
                )
            if not problem0[MODEL_FILES]:
                model = None
            else:
                model_id, model_info = next(
                    iter(problem0[MODEL_FILES].items())
                )
                model = model_factory(
                    get_path(model_info[MODEL_LOCATION]),
                    model_info[MODEL_LANGUAGE],
                    model_id=model_id,
                )

        measurement_files = [
            get_path(f) for f in problem0.get(MEASUREMENT_FILES, [])
        ]
        # If there are multiple tables, we will merge them
        measurement_df = (
            core.concat_tables(
                measurement_files, measurements.get_measurement_df
            )
            if measurement_files
            else None
        )

        condition_files = [
            get_path(f) for f in problem0.get(CONDITION_FILES, [])
        ]
        # If there are multiple tables, we will merge them
        condition_df = (
            core.concat_tables(condition_files, conditions.get_condition_df)
            if condition_files
            else None
        )

        visualization_files = [
            get_path(f) for f in problem0.get(VISUALIZATION_FILES, [])
        ]
        # If there are multiple tables, we will merge them
        visualization_df = (
            core.concat_tables(visualization_files, core.get_visualization_df)
            if visualization_files
            else None
        )

        observable_files = [
            get_path(f) for f in problem0.get(OBSERVABLE_FILES, [])
        ]
        # If there are multiple tables, we will merge them
        observable_df = (
            core.concat_tables(observable_files, observables.get_observable_df)
            if observable_files
            else None
        )

        mapping_files = [get_path(f) for f in problem0.get(MAPPING_FILES, [])]
        # If there are multiple tables, we will merge them
        mapping_df = (
            core.concat_tables(mapping_files, mapping.get_mapping_df)
            if mapping_files
            else None
        )

        return Problem(
            condition_df=condition_df,
            measurement_df=measurement_df,
            parameter_df=parameter_df,
            observable_df=observable_df,
            model=model,
            visualization_df=visualization_df,
            mapping_df=mapping_df,
            extensions_config=yaml_config.get(EXTENSIONS, {}),
        )

    @staticmethod
    def from_combine(filename: Union[Path, str]) -> "Problem":
        """Read PEtab COMBINE archive (http://co.mbine.org/documents/archive).

        See also :py:func:`petab.create_combine_archive`.

        Arguments:
            filename: Path to the PEtab-COMBINE archive

        Returns:
            A :py:class:`petab.Problem` instance.
        """
        # function-level import, because module-level import interfered with
        # other SWIG interfaces
        try:
            import libcombine
        except ImportError as e:
            raise ImportError(
                "To use PEtab's COMBINE functionality, libcombine "
                "(python-libcombine) must be installed."
            ) from e

        archive = libcombine.CombineArchive()
        if archive.initializeFromArchive(str(filename)) is None:
            print(f"Invalid Combine Archive: {filename}")
            return None

        with tempfile.TemporaryDirectory() as tmpdirname:
            archive.extractTo(tmpdirname)
            problem = Problem.from_yaml(
                os.path.join(tmpdirname, archive.getMasterFile().getLocation())
            )
        archive.cleanUp()

        return problem

    def to_files_generic(
        self,
        prefix_path: Union[str, Path],
    ) -> str:
        """Save a PEtab problem to generic file names.

        The PEtab problem YAML file is always created. PEtab data files are
        only created if the PEtab problem contains corresponding data (e.g. a
        PEtab visualization TSV file is only created if the PEtab problem has
        one).

        Arguments:
            prefix_path:
                Specify a prefix to all paths, to avoid specifying the
                prefix for all paths individually. NB: the prefix is added to
                paths before ``relative_paths`` is handled downstream in
                :func:`petab.yaml.create_problem_yaml`.

        Returns:
            The path to the PEtab problem YAML file.
        """
        prefix_path = Path(prefix_path)

        # Generate generic filenames for data tables in the PEtab problem that
        # contain data.
        filenames = {}
        for table_name in [
            "condition",
            "measurement",
            "parameter",
            "observable",
            "visualization",
            "mapping",
        ]:
            if getattr(self, f"{table_name}_df") is not None:
                filenames[f"{table_name}_file"] = f"{table_name}s.tsv"

        if self.model:
            if not isinstance(self.model, SbmlModel):
                raise NotImplementedError(
                    "Saving non-SBML models is " "currently not supported."
                )
            filenames["model_file"] = "model.xml"

        filenames["yaml_file"] = "problem.yaml"

        self.to_files(**filenames, prefix_path=prefix_path)

        if prefix_path is None:
            return filenames["yaml_file"]
        return str(prefix_path / filenames["yaml_file"])

    def to_files(
        self,
        sbml_file: Union[None, str, Path] = None,
        condition_file: Union[None, str, Path] = None,
        measurement_file: Union[None, str, Path] = None,
        parameter_file: Union[None, str, Path] = None,
        visualization_file: Union[None, str, Path] = None,
        observable_file: Union[None, str, Path] = None,
        yaml_file: Union[None, str, Path] = None,
        prefix_path: Union[None, str, Path] = None,
        relative_paths: bool = True,
        model_file: Union[None, str, Path] = None,
        mapping_file: Union[None, str, Path] = None,
    ) -> None:
        """
        Write PEtab tables to files for this problem

        Writes PEtab files for those entities for which a destination was
        passed.

        NOTE: If this instance was created from multiple measurement or
        visualization tables, they will be merged and written to a single file.

        Arguments:
            sbml_file: SBML model destination (deprecated)
            model_file: Model destination
            condition_file: Condition table destination
            measurement_file: Measurement table destination
            parameter_file: Parameter table destination
            visualization_file: Visualization table destination
            observable_file: Observables table destination
            mapping_file: Mapping table destination
            yaml_file: YAML file destination
            prefix_path:
                Specify a prefix to all paths, to avoid specifying the
                prefix for all paths individually. NB: the prefix is added to
                paths before ``relative_paths`` is handled.
            relative_paths:
                whether all paths in the YAML file should be
                relative to the location of the YAML file. If ``False``, then
                paths are left unchanged.

        Raises:
            ValueError:
                If a destination was provided for a non-existing entity.
        """
        if sbml_file:
            warn(
                "The `sbml_file` argument is deprecated and will be "
                "removed in a future version. Use `model_file` instead.",
                DeprecationWarning,
                stacklevel=2,
            )

            if model_file:
                raise ValueError(
                    "Must provide either `sbml_file` or "
                    "`model_file` argument, but not both."
                )

            model_file = sbml_file

        if prefix_path is not None:
            prefix_path = Path(prefix_path)

            def add_prefix(path0: Union[None, str, Path]) -> str:
                return path0 if path0 is None else str(prefix_path / path0)

            model_file = add_prefix(model_file)
            condition_file = add_prefix(condition_file)
            measurement_file = add_prefix(measurement_file)
            parameter_file = add_prefix(parameter_file)
            observable_file = add_prefix(observable_file)
            visualization_file = add_prefix(visualization_file)
            mapping_file = add_prefix(mapping_file)
            yaml_file = add_prefix(yaml_file)

        if model_file:
            self.model.to_file(model_file)

        def error(name: str) -> ValueError:
            return ValueError(f"Unable to save non-existent {name} table")

        if condition_file:
            if self.condition_df is not None:
                conditions.write_condition_df(
                    self.condition_df, condition_file
                )
            else:
                raise error("condition")

        if measurement_file:
            if self.measurement_df is not None:
                measurements.write_measurement_df(
                    self.measurement_df, measurement_file
                )
            else:
                raise error("measurement")

        if parameter_file:
            if self.parameter_df is not None:
                parameters.write_parameter_df(
                    self.parameter_df, parameter_file
                )
            else:
                raise error("parameter")

        if observable_file:
            if self.observable_df is not None:
                observables.write_observable_df(
                    self.observable_df, observable_file
                )
            else:
                raise error("observable")

        if visualization_file:
            if self.visualization_df is not None:
                core.write_visualization_df(
                    self.visualization_df, visualization_file
                )
            else:
                raise error("visualization")

        if mapping_file:
            if self.mapping_df is not None:
                mapping.write_mapping_df(self.mapping_df, mapping_file)
            else:
                raise error("mapping")

        if yaml_file:
            yaml.create_problem_yaml(
                sbml_files=model_file,
                condition_files=condition_file,
                measurement_files=measurement_file,
                parameter_file=parameter_file,
                observable_files=observable_file,
                yaml_file=yaml_file,
                visualization_files=visualization_file,
                relative_paths=relative_paths,
                mapping_files=mapping_file,
            )

    def get_optimization_parameters(self):
        """
        Return list of optimization parameter IDs.

        See :py:func:`petab.parameters.get_optimization_parameters`.
        """
        return parameters.get_optimization_parameters(self.parameter_df)

    def get_optimization_parameter_scales(self):
        """
        Return list of optimization parameter scaling strings.

        See :py:func:`petab.parameters.get_optimization_parameters`.
        """
        return parameters.get_optimization_parameter_scaling(self.parameter_df)

    def get_model_parameters(self):
        """See :py:func:`petab.sbml.get_model_parameters`"""
        warn(
            "petab.Problem.get_model_parameters is deprecated and will be "
            "removed in a future version.",
            DeprecationWarning,
            stacklevel=2,
        )

        return sbml.get_model_parameters(self.sbml_model)

    def get_observable_ids(self):
        """
        Returns dictionary of observable ids.
        """
        return list(self.observable_df.index)

    def _apply_mask(self, v: List, free: bool = True, fixed: bool = True):
        """Apply mask of only free or only fixed values.

        Parameters
        ----------
        v:
            The full vector the mask is to be applied to.
        free:
            Whether to return free parameters, i.e. parameters to estimate.
        fixed:
            Whether to return fixed parameters, i.e. parameters not to
            estimate.

        Returns
        -------
        The reduced vector with applied mask.
        """
        if not free and not fixed:
            return []
        if not free:
            return [v[ix] for ix in self.x_fixed_indices]
        if not fixed:
            return [v[ix] for ix in self.x_free_indices]
        return v

    def get_x_ids(self, free: bool = True, fixed: bool = True):
        """Generic function to get parameter ids.

        Parameters
        ----------
        free:
            Whether to return free parameters, i.e. parameters to estimate.
        fixed:
            Whether to return fixed parameters, i.e. parameters not to
            estimate.

        Returns
        -------
        The parameter IDs.
        """
        v = list(self.parameter_df.index.values)
        return self._apply_mask(v, free=free, fixed=fixed)

    @property
    def x_ids(self) -> List[str]:
        """Parameter table parameter IDs"""
        return self.get_x_ids()

    @property
    def x_free_ids(self) -> List[str]:
        """Parameter table parameter IDs, for free parameters."""
        return self.get_x_ids(fixed=False)

    @property
    def x_fixed_ids(self) -> List[str]:
        """Parameter table parameter IDs, for fixed parameters."""
        return self.get_x_ids(free=False)

    def get_x_nominal(
        self, free: bool = True, fixed: bool = True, scaled: bool = False
    ):
        """Generic function to get parameter nominal values.

        Parameters
        ----------
        free:
            Whether to return free parameters, i.e. parameters to estimate.
        fixed:
            Whether to return fixed parameters, i.e. parameters not to
            estimate.
        scaled:
            Whether to scale the values according to the parameter scale,
            or return them on linear scale.

        Returns
        -------
        The parameter nominal values.
        """
        if NOMINAL_VALUE in self.parameter_df:
            v = list(self.parameter_df[NOMINAL_VALUE])
        else:
            v = [nan] * len(self.parameter_df)

        if scaled:
            v = list(
                parameters.map_scale(v, self.parameter_df[PARAMETER_SCALE])
            )
        return self._apply_mask(v, free=free, fixed=fixed)

    @property
    def x_nominal(self) -> List:
        """Parameter table nominal values"""
        return self.get_x_nominal()

    @property
    def x_nominal_free(self) -> List:
        """Parameter table nominal values, for free parameters."""
        return self.get_x_nominal(fixed=False)

    @property
    def x_nominal_fixed(self) -> List:
        """Parameter table nominal values, for fixed parameters."""
        return self.get_x_nominal(free=False)

    @property
    def x_nominal_scaled(self) -> List:
        """Parameter table nominal values with applied parameter scaling"""
        return self.get_x_nominal(scaled=True)

    @property
    def x_nominal_free_scaled(self) -> List:
        """Parameter table nominal values with applied parameter scaling,
        for free parameters."""
        return self.get_x_nominal(fixed=False, scaled=True)

    @property
    def x_nominal_fixed_scaled(self) -> List:
        """Parameter table nominal values with applied parameter scaling,
        for fixed parameters."""
        return self.get_x_nominal(free=False, scaled=True)

    def get_lb(
        self, free: bool = True, fixed: bool = True, scaled: bool = False
    ):
        """Generic function to get lower parameter bounds.

        Parameters
        ----------
        free:
            Whether to return free parameters, i.e. parameters to estimate.
        fixed:
            Whether to return fixed parameters, i.e. parameters not to
            estimate.
        scaled:
            Whether to scale the values according to the parameter scale,
            or return them on linear scale.

        Returns
        -------
        The lower parameter bounds.
        """
        v = list(self.parameter_df[LOWER_BOUND])
        if scaled:
            v = list(
                parameters.map_scale(v, self.parameter_df[PARAMETER_SCALE])
            )
        return self._apply_mask(v, free=free, fixed=fixed)

    @property
    def lb(self) -> List:
        """Parameter table lower bounds."""
        return self.get_lb()

    @property
    def lb_scaled(self) -> List:
        """Parameter table lower bounds with applied parameter scaling"""
        return self.get_lb(scaled=True)

    def get_ub(
        self, free: bool = True, fixed: bool = True, scaled: bool = False
    ):
        """Generic function to get upper parameter bounds.

        Parameters
        ----------
        free:
            Whether to return free parameters, i.e. parameters to estimate.
        fixed:
            Whether to return fixed parameters, i.e. parameters not to
            estimate.
        scaled:
            Whether to scale the values according to the parameter scale,
            or return them on linear scale.

        Returns
        -------
        The upper parameter bounds.
        """
        v = list(self.parameter_df[UPPER_BOUND])
        if scaled:
            v = list(
                parameters.map_scale(v, self.parameter_df[PARAMETER_SCALE])
            )
        return self._apply_mask(v, free=free, fixed=fixed)

    @property
    def ub(self) -> List:
        """Parameter table upper bounds"""
        return self.get_ub()

    @property
    def ub_scaled(self) -> List:
        """Parameter table upper bounds with applied parameter scaling"""
        return self.get_ub(scaled=True)

    @property
    def x_free_indices(self) -> List[int]:
        """Parameter table estimated parameter indices."""
        estimated = list(self.parameter_df[ESTIMATE])
        return [j for j, val in enumerate(estimated) if val != 0]

    @property
    def x_fixed_indices(self) -> List[int]:
        """Parameter table non-estimated parameter indices."""
        estimated = list(self.parameter_df[ESTIMATE])
        return [j for j, val in enumerate(estimated) if val == 0]

    def get_simulation_conditions_from_measurement_df(self):
        """See petab.get_simulation_conditions"""
        return measurements.get_simulation_conditions(self.measurement_df)

    def get_optimization_to_simulation_parameter_mapping(self, **kwargs):
        """
        See
        :py:func:`petab.parameter_mapping.get_optimization_to_simulation_parameter_mapping`,
        to which all keyword arguments are forwarded.
        """
        return (
            parameter_mapping.get_optimization_to_simulation_parameter_mapping(
                condition_df=self.condition_df,
                measurement_df=self.measurement_df,
                parameter_df=self.parameter_df,
                observable_df=self.observable_df,
                model=self.model,
                **kwargs,
            )
        )

    def create_parameter_df(self, *args, **kwargs):
        """Create a new PEtab parameter table

        See :py:func:`create_parameter_df`.
        """
        return parameters.create_parameter_df(
            model=self.model,
            condition_df=self.condition_df,
            observable_df=self.observable_df,
            measurement_df=self.measurement_df,
            mapping_df=self.mapping_df,
            *args,
            **kwargs,
        )

    def sample_parameter_startpoints(self, n_starts: int = 100, **kwargs):
        """Create 2D array with starting points for optimization

        See :py:func:`petab.sample_parameter_startpoints`.
        """
        return sampling.sample_parameter_startpoints(
            self.parameter_df, n_starts=n_starts, **kwargs
        )

    def sample_parameter_startpoints_dict(
        self, n_starts: int = 100
    ) -> List[Dict[str, float]]:
        """Create dictionaries with starting points for optimization

        See also :py:func:`petab.sample_parameter_startpoints`.

        Returns:
            A list of dictionaries with parameter IDs mapping to samples
            parameter values.
        """
        return [
            dict(zip(self.x_free_ids, parameter_values))
            for parameter_values in self.sample_parameter_startpoints(
                n_starts=n_starts
            )
        ]

    def unscale_parameters(
        self,
        x_dict: Dict[str, float],
    ) -> Dict[str, float]:
        """Unscale parameter values.

        Parameters
        ----------
        x_dict:
            Keys are parameter IDs in the PEtab problem, values are scaled
            parameter values.

        Returns
        -------
        The unscaled parameter values.
        """
        return {
            parameter_id: parameters.unscale(
                parameter_value,
                self.parameter_df[PARAMETER_SCALE][parameter_id],
            )
            for parameter_id, parameter_value in x_dict.items()
        }

    def scale_parameters(
        self,
        x_dict: Dict[str, float],
    ) -> Dict[str, float]:
        """Scale parameter values.

        Parameters
        ----------
        x_dict:
            Keys are parameter IDs in the PEtab problem, values are unscaled
            parameter values.

        Returns
        -------
        The scaled parameter values.
        """
        return {
            parameter_id: parameters.scale(
                parameter_value,
                self.parameter_df[PARAMETER_SCALE][parameter_id],
            )
            for parameter_id, parameter_value in x_dict.items()
        }

    @property
    def n_estimated(self) -> int:
        """The number of estimated parameters."""
        return len(self.x_free_indices)

    @property
    def n_measurements(self) -> int:
        """Number of measurements."""
        return self.measurement_df[MEASUREMENT].notna().sum()

    @property
    def n_priors(self) -> int:
        """Number of priors."""
        if OBJECTIVE_PRIOR_PARAMETERS not in self.parameter_df:
            return 0

        return self.parameter_df[OBJECTIVE_PRIOR_PARAMETERS].notna().sum()



"""Tests related to petab.measurements"""
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd

import petab
from petab.C import *


def test_get_measurement_df():
    """Test measurements.get_measurement_df."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        measurement_df.to_csv(fh, sep="\t", index=False)

    df = petab.get_measurement_df(file_name).replace(np.nan, "")
    assert (df == measurement_df).all().all()

    # test other arguments
    assert (
        (petab.get_measurement_df(measurement_df) == measurement_df)
        .all()
        .all()
    )
    assert petab.get_measurement_df(None) is None


def test_write_measurement_df():
    """Test measurements.get_measurement_df."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )

    with tempfile.TemporaryDirectory() as temp_dir:
        file_name = Path(temp_dir) / "parameters.tsv"
        petab.write_measurement_df(measurement_df, file_name)
        re_df = petab.get_measurement_df(file_name).replace(np.nan, "")
        assert (measurement_df == re_df).all().all()


def test_create_measurement_df():
    """Test measurements.create_measurement_df."""
    df = petab.create_measurement_df()
    assert set(df.columns.values) == set(MEASUREMENT_DF_COLS)


def test_measurements_have_replicates():
    """Test measurements.measurements_have_replicates."""
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs1"],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
            TIME: [0, 1],
            MEASUREMENT: [42, 137.01],
        }
    )
    assert not petab.measurements_have_replicates(measurement_df)

    measurement_df[TIME] = [1, 1]
    assert petab.measurements_have_replicates(measurement_df)


def test_get_simulation_conditions():
    """Test get_simulation_conditions"""

    # only simulation condition
    measurement_df = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
        }
    )
    expected = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1"],
        }
    )
    actual = petab.get_simulation_conditions(measurement_df)
    assert actual.equals(expected)

    # simulation and preequilibration condition
    measurement_df = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: ["c1", "c0", "c1", "c0"],
        }
    )
    expected = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: ["c1", "c0"],
        }
    )
    actual = petab.get_simulation_conditions(measurement_df)
    assert actual.equals(expected)

    # simulation with and without preequilibration
    measurement_df = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: ["", "", "c1", "c0"],
        }
    )
    expected = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: ["", "", "c1", "c0"],
        }
    ).sort_values(
        [SIMULATION_CONDITION_ID, PREEQUILIBRATION_CONDITION_ID],
        ignore_index=True,
    )
    actual = petab.get_simulation_conditions(measurement_df)
    assert actual.equals(expected)

    # simulation with and without preequilibration; NaNs
    measurement_df = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: [np.nan, np.nan, "c1", "c0"],
        }
    )
    expected = pd.DataFrame(
        data={
            SIMULATION_CONDITION_ID: ["c0", "c1", "c0", "c1"],
            PREEQUILIBRATION_CONDITION_ID: ["", "", "c1", "c0"],
        }
    ).sort_values(
        [SIMULATION_CONDITION_ID, PREEQUILIBRATION_CONDITION_ID],
        ignore_index=True,
    )
    actual = petab.get_simulation_conditions(measurement_df)
    assert actual.equals(expected)



"""Test for petab.yaml"""
import tempfile
from pathlib import Path

import pytest
from jsonschema.exceptions import ValidationError

from petab.yaml import create_problem_yaml, validate


def test_validate():
    data = {"format_version": "1"}

    # should fail because we miss some information
    with pytest.raises(ValidationError):
        validate(data)

    # should be well-formed
    file_ = (
        Path(__file__).parents[1]
        / "doc"
        / "example"
        / "example_Fujita"
        / "Fujita.yaml"
    )
    validate(file_)


def test_create_problem_yaml():
    with tempfile.TemporaryDirectory() as outdir:
        # test with single problem files
        # create target files
        sbml_file = Path(outdir, "model.xml")
        condition_file = Path(outdir, "conditions.tsv")
        measurement_file = Path(outdir, "measurements.tsv")
        parameter_file = Path(outdir, "parameters.tsv")
        observable_file = Path(outdir, "observables.tsv")
        yaml_file = Path(outdir, "problem.yaml")
        visualization_file = Path(outdir, "visualization.tsv")
        for file in (
            sbml_file,
            condition_file,
            measurement_file,
            parameter_file,
            observable_file,
            visualization_file,
        ):
            file.touch()
        create_problem_yaml(
            sbml_file,
            condition_file,
            measurement_file,
            parameter_file,
            observable_file,
            yaml_file,
            visualization_file,
        )
        validate(yaml_file)

        # test for list of files
        # create additional target files
        sbml_file2 = Path(outdir, "model2.xml")
        condition_file2 = Path(outdir, "conditions2.tsv")
        measurement_file2 = Path(outdir, "measurements2.tsv")
        observable_file2 = Path(outdir, "observables2.tsv")
        yaml_file2 = Path(outdir, "problem2.yaml")
        for file in (
            sbml_file2,
            condition_file2,
            measurement_file2,
            observable_file2,
        ):
            file.touch()

        sbml_files = [sbml_file, sbml_file2]
        condition_files = [condition_file, condition_file2]
        measurement_files = [measurement_file, measurement_file2]
        observable_files = [observable_file, observable_file2]
        create_problem_yaml(
            sbml_files,
            condition_files,
            measurement_files,
            parameter_file,
            observable_files,
            yaml_file2,
        )
        validate(yaml_file2)



import copy
import pickle
import tempfile
import warnings
from io import StringIO
from math import nan
from pathlib import Path

import libsbml
import numpy as np
import pandas as pd
import pytest
from yaml import safe_load

import petab
from petab.C import *
from petab.models.sbml_model import SbmlModel


@pytest.fixture
def condition_df_2_conditions():
    condition_df = pd.DataFrame(
        data={
            "conditionId": ["condition1", "condition2"],
            "conditionName": ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )
    condition_df.set_index("conditionId", inplace=True)
    condition_df.fixedParameter1 = condition_df.fixedParameter1.astype(
        "object"
    )
    return condition_df


@pytest.fixture
def petab_problem():
    """Test petab problem."""
    # create test model
    import simplesbml

    model = simplesbml.SbmlModel()
    model.addParameter("fixedParameter1", 0.0)
    model.addParameter("observable_1", 0.0)

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            MEASUREMENT: [0.1, 0.2],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )

    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    ).set_index(CONDITION_ID)

    parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["dynamicParameter1", "dynamicParameter2"],
            PARAMETER_NAME: ["", "..."],
            ESTIMATE: [1, 0],
        }
    ).set_index(PARAMETER_ID)

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1"],
            OBSERVABLE_NAME: ["julius"],
            OBSERVABLE_FORMULA: ["observable_1 * observableParameter1_obs1"],
            NOISE_FORMULA: ["0.1 * observable_1 * observableParameter1_obs1"],
        }
    ).set_index(OBSERVABLE_ID)

    with tempfile.TemporaryDirectory() as temp_dir:
        sbml_file_name = Path(temp_dir, "model.xml")
        libsbml.writeSBMLToFile(model.document, str(sbml_file_name))

        measurement_file_name = Path(temp_dir, "measurements.tsv")
        petab.write_measurement_df(measurement_df, measurement_file_name)

        condition_file_name = Path(temp_dir, "conditions.tsv")
        petab.write_condition_df(condition_df, condition_file_name)

        parameter_file_name = Path(temp_dir, "parameters.tsv")
        petab.write_parameter_df(parameter_df, parameter_file_name)

        observable_file_name = Path(temp_dir, "observables.tsv")
        petab.write_observable_df(observable_df, observable_file_name)

        with pytest.deprecated_call():
            petab_problem = petab.Problem.from_files(
                sbml_file=sbml_file_name,
                measurement_file=measurement_file_name,
                condition_file=condition_file_name,
                parameter_file=parameter_file_name,
                observable_files=observable_file_name,
            )
            assert petab_problem.n_measurements == 2
            assert petab_problem.n_estimated == 1
            assert petab_problem.n_priors == 0

            yield petab_problem


@pytest.fixture
def fujita_model_scaling():
    path = Path(__file__).parent.parent / "doc" / "example" / "example_Fujita"

    sbml_file = path / "Fujita_model.xml"
    condition_file = path / "Fujita_experimentalCondition.tsv"
    measurement_file = path / "Fujita_measurementData.tsv"
    parameter_file = path / "Fujita_parameters_scaling.tsv"

    with pytest.deprecated_call():
        return petab.Problem.from_files(
            sbml_file=sbml_file,
            condition_file=condition_file,
            measurement_file=measurement_file,
            parameter_file=parameter_file,
        )


def test_split_parameter_replacement_list():
    assert petab.split_parameter_replacement_list("") == []
    assert petab.split_parameter_replacement_list("param1") == ["param1"]
    assert petab.split_parameter_replacement_list("param1;param2") == [
        "param1",
        "param2",
    ]
    assert petab.split_parameter_replacement_list("1.0") == [1.0]
    assert petab.split_parameter_replacement_list("1.0;2.0") == [1.0, 2.0]
    assert petab.split_parameter_replacement_list("param1;2.2") == [
        "param1",
        2.2,
    ]
    assert petab.split_parameter_replacement_list(np.nan) == []
    assert petab.split_parameter_replacement_list(1.5) == [1.5]
    assert petab.split_parameter_replacement_list(None) == []

    with pytest.raises(ValueError):
        assert petab.split_parameter_replacement_list("1.0;")

    with pytest.raises(ValueError):
        assert petab.split_parameter_replacement_list(";1.0")


def test_get_measurement_parameter_ids():
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )
    expected = ["p1", "p2", "p3", "p4", "p5"]
    actual = petab.get_measurement_parameter_ids(measurement_df)
    # ordering is arbitrary
    assert set(actual) == set(expected)


def test_serialization(petab_problem):
    # serialize and back
    problem_recreated = pickle.loads(pickle.dumps(petab_problem))

    assert problem_recreated.measurement_df.equals(
        petab_problem.measurement_df
    )

    assert problem_recreated.parameter_df.equals(petab_problem.parameter_df)

    assert problem_recreated.condition_df.equals(petab_problem.condition_df)

    # Can't test for equality directly, testing for number of parameters
    #  should do the job here
    assert len(problem_recreated.sbml_model.getListOfParameters()) == len(
        petab_problem.sbml_model.getListOfParameters()
    )


def test_get_priors_from_df():
    """Check petab.get_priors_from_df."""
    parameter_df = pd.DataFrame(
        {
            PARAMETER_ID: ["p1", "p2", "p3", "p4", "p5"],
            PARAMETER_SCALE: [LOG10, LOG10, LOG10, LOG10, LOG10],
            LOWER_BOUND: [1e-8, 1e-9, 1e-10, 1e-11, 1e-5],
            UPPER_BOUND: [1e8, 1e9, 1e10, 1e11, 1e5],
            ESTIMATE: [1, 1, 1, 1, 0],
            INITIALIZATION_PRIOR_TYPE: ["", "", UNIFORM, NORMAL, ""],
            INITIALIZATION_PRIOR_PARAMETERS: [
                "",
                "-5;5",
                "1e-5;1e5",
                "0;1",
                "",
            ],
        }
    )
    parameter_df = petab.get_parameter_df(parameter_df)

    prior_list = petab.get_priors_from_df(parameter_df, mode=INITIALIZATION)

    # only give values for estimated parameters
    assert len(prior_list) == 4

    # correct types
    types = [entry[0] for entry in prior_list]
    assert types == [
        PARAMETER_SCALE_UNIFORM,
        PARAMETER_SCALE_UNIFORM,
        UNIFORM,
        NORMAL,
    ]

    # correct scales
    scales = [entry[2] for entry in prior_list]
    assert scales == [LOG10] * 4

    # correct bounds
    bounds = [entry[3] for entry in prior_list]
    assert (
        bounds
        == list(zip(parameter_df[LOWER_BOUND], parameter_df[UPPER_BOUND]))[:4]
    )

    # give correct value for empty
    prior_pars = [entry[1] for entry in prior_list]
    assert prior_pars[0] == (-8, 8)
    assert prior_pars[1] == (-5, 5)
    assert prior_pars[2] == (1e-5, 1e5)

    # check subsetting / reordering works
    prior_list_subset = petab.get_priors_from_df(
        parameter_df, mode=INITIALIZATION, parameter_ids=["p2", "p1"]
    )
    assert len(prior_list_subset) == 2
    assert prior_list_subset == [prior_list[1], prior_list[0]]

    with pytest.raises(KeyError, match="Parameter table does not contain"):
        petab.get_priors_from_df(
            parameter_df, mode=INITIALIZATION, parameter_ids=["non_existent"]
        )


def test_startpoint_sampling(fujita_model_scaling):
    n_starts = 10
    startpoints = fujita_model_scaling.sample_parameter_startpoints(n_starts)
    assert (np.isfinite(startpoints)).all
    assert startpoints.shape == (n_starts, 19)
    for sp in startpoints:
        assert np.log10(31.62) <= sp[0] <= np.log10(316.23)
        assert -3 <= sp[1] <= 3


def test_startpoint_sampling_dict(fujita_model_scaling):
    n_starts = 10
    startpoints = fujita_model_scaling.sample_parameter_startpoints_dict(
        n_starts
    )
    assert len(startpoints) == n_starts
    for startpoint in startpoints:
        assert set(startpoint.keys()) == set(fujita_model_scaling.x_free_ids)


def test_create_parameter_df(
    condition_df_2_conditions,
):  # pylint: disable=W0621
    """Test petab.create_parameter_df."""
    import simplesbml

    ss_model = simplesbml.SbmlModel()
    ss_model.addSpecies("[x1]", 1.0)
    ss_model.addParameter("fixedParameter1", 2.0)
    ss_model.addParameter("p0", 3.0)
    model = SbmlModel(sbml_model=ss_model.model)

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_FORMULA: ["x1", "2*x1"],
        }
    ).set_index(OBSERVABLE_ID)

    # Add assignment rule target which should be ignored
    ss_model.addParameter("assignment_target", 0.0)
    ss_model.addAssignmentRule("assignment_target", "1.0")

    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )

    # first model parameters, then row by row noise and sigma overrides
    expected = ["p3", "p4", "p1", "p2", "p5"]

    # Test old API with passing libsbml.Model directly
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        parameter_df = petab.create_parameter_df(
            ss_model.model,
            condition_df_2_conditions,
            observable_df,
            measurement_df,
        )
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        assert parameter_df.index.values.tolist() == expected

    parameter_df = petab.create_parameter_df(
        model=model,
        condition_df=condition_df_2_conditions,
        observable_df=observable_df,
        measurement_df=measurement_df,
    )
    assert parameter_df.index.values.tolist() == expected

    # test with condition parameter override:
    condition_df_2_conditions.loc[
        "condition2", "fixedParameter1"
    ] = "overrider"
    expected = ["p3", "p4", "p1", "p2", "p5", "overrider"]

    parameter_df = petab.create_parameter_df(
        model=model,
        condition_df=condition_df_2_conditions,
        observable_df=observable_df,
        measurement_df=measurement_df,
    )
    actual = parameter_df.index.values.tolist()
    assert actual == expected

    # test with optional parameters
    expected = ["p0", "p3", "p4", "p1", "p2", "p5", "overrider"]

    parameter_df = petab.create_parameter_df(
        model=model,
        condition_df=condition_df_2_conditions,
        observable_df=observable_df,
        measurement_df=measurement_df,
        include_optional=True,
    )
    actual = parameter_df.index.values.tolist()
    assert actual == expected
    assert parameter_df.loc["p0", NOMINAL_VALUE] == 3.0


def test_flatten_timepoint_specific_output_overrides():
    """Test flatten_timepoint_specific_output_overrides"""
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_FORMULA: [
                "observableParameter1_obs1 + observableParameter2_obs1",
                "x",
            ],
            NOISE_FORMULA: [
                "(observableParameter1_obs1 + observableParameter2_obs1) * noiseParameter1_obs1",
                1,
            ],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    # new observable IDs (obs${i_obs}_${i_obsParOverride}_${i_noiseParOverride}_${i_condition})
    obs1_1_1_1 = "obs1__obsParOverride1_1_0__noiseParOverride1__condition1"
    obs1_2_1_1 = "obs1__obsParOverride2_1_0__noiseParOverride1__condition1"
    obs1_2_2_1 = "obs1__obsParOverride2_1_0__noiseParOverride2__condition1"
    observable_df_expected = pd.DataFrame(
        data={
            OBSERVABLE_ID: [
                obs1_1_1_1,
                obs1_2_1_1,
                obs1_2_2_1,
                "obs2__condition1",
            ],
            OBSERVABLE_FORMULA: [
                f"observableParameter1_{obs1_1_1_1} + observableParameter2_{obs1_1_1_1}",
                f"observableParameter1_{obs1_2_1_1} + observableParameter2_{obs1_2_1_1}",
                f"observableParameter1_{obs1_2_2_1} + observableParameter2_{obs1_2_2_1}",
                "x",
            ],
            NOISE_FORMULA: [
                f"(observableParameter1_{obs1_1_1_1} + observableParameter2_{obs1_1_1_1})"
                f" * noiseParameter1_{obs1_1_1_1}",
                f"(observableParameter1_{obs1_2_1_1} + observableParameter2_{obs1_2_1_1})"
                f" * noiseParameter1_{obs1_2_1_1}",
                f"(observableParameter1_{obs1_2_2_1} + observableParameter2_{obs1_2_2_1})"
                f" * noiseParameter1_{obs1_2_2_1}",
                1,
            ],
        }
    )
    observable_df_expected.set_index(OBSERVABLE_ID, inplace=True)

    # Measurement table with timepoint-specific overrides
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs1", "obs1", "obs1", "obs2"],
            SIMULATION_CONDITION_ID: [
                "condition1",
                "condition1",
                "condition1",
                "condition1",
                "condition1",
            ],
            PREEQUILIBRATION_CONDITION_ID: ["", "", "", "", ""],
            TIME: [1.0, 1.0, 2.0, 2.0, 3.0],
            MEASUREMENT: [0.1] * 5,
            OBSERVABLE_PARAMETERS: [
                "obsParOverride1;1.0",
                "obsParOverride2;1.0",
                "obsParOverride2;1.0",
                "obsParOverride2;1.0",
                "",
            ],
            NOISE_PARAMETERS: [
                "noiseParOverride1",
                "noiseParOverride1",
                "noiseParOverride2",
                "noiseParOverride2",
                "",
            ],
        }
    )

    measurement_df_expected = pd.DataFrame(
        data={
            OBSERVABLE_ID: [
                obs1_1_1_1,
                obs1_2_1_1,
                obs1_2_2_1,
                obs1_2_2_1,
                "obs2__condition1",
            ],
            SIMULATION_CONDITION_ID: [
                "condition1",
                "condition1",
                "condition1",
                "condition1",
                "condition1",
            ],
            PREEQUILIBRATION_CONDITION_ID: ["", "", "", "", ""],
            TIME: [1.0, 1.0, 2.0, 2.0, 3.0],
            MEASUREMENT: [0.1] * 5,
            OBSERVABLE_PARAMETERS: [
                "obsParOverride1;1.0",
                "obsParOverride2;1.0",
                "obsParOverride2;1.0",
                "obsParOverride2;1.0",
                "",
            ],
            NOISE_PARAMETERS: [
                "noiseParOverride1",
                "noiseParOverride1",
                "noiseParOverride2",
                "noiseParOverride2",
                "",
            ],
        }
    )

    problem = petab.Problem(
        measurement_df=measurement_df, observable_df=observable_df
    )

    assert petab.lint_problem(problem) is False

    # Ensure having timepoint-specific overrides
    assert (
        petab.lint.measurement_table_has_timepoint_specific_mappings(
            measurement_df
        )
        is True
    )

    petab.flatten_timepoint_specific_output_overrides(problem)

    # Timepoint-specific overrides should be gone now
    assert (
        petab.lint.measurement_table_has_timepoint_specific_mappings(
            problem.measurement_df
        )
        is False
    )

    pd.testing.assert_frame_equal(
        problem.observable_df, observable_df_expected
    )
    pd.testing.assert_frame_equal(
        problem.measurement_df, measurement_df_expected
    )

    assert petab.lint_problem(problem) is False

    simulation_df = copy.deepcopy(problem.measurement_df)
    simulation_df.rename(columns={MEASUREMENT: SIMULATION})
    unflattened_problem = petab.Problem(
        measurement_df=measurement_df,
        observable_df=observable_df,
    )
    unflattened_simulation_df = petab.core.unflatten_simulation_df(
        simulation_df=simulation_df,
        petab_problem=unflattened_problem,
    )
    # The unflattened simulation dataframe has the original observable IDs.
    assert (
        unflattened_simulation_df[OBSERVABLE_ID] == ["obs1"] * 4 + ["obs2"]
    ).all()


def test_flatten_timepoint_specific_output_overrides_special_cases():
    """Test flatten_timepoint_specific_output_overrides
    for special cases:
    * no preequilibration
    * no observable parameters
    """
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1"],
            OBSERVABLE_FORMULA: ["species1"],
            NOISE_FORMULA: ["noiseParameter1_obs1"],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    observable_df_expected = pd.DataFrame(
        data={
            OBSERVABLE_ID: [
                "obs1__noiseParOverride1__condition1",
                "obs1__noiseParOverride2__condition1",
            ],
            OBSERVABLE_FORMULA: ["species1", "species1"],
            NOISE_FORMULA: [
                "noiseParameter1_obs1__noiseParOverride1__condition1",
                "noiseParameter1_obs1__noiseParOverride2__condition1",
            ],
        }
    )
    observable_df_expected.set_index(OBSERVABLE_ID, inplace=True)

    # Measurement table with timepoint-specific overrides
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs1", "obs1", "obs1"],
            SIMULATION_CONDITION_ID: [
                "condition1",
                "condition1",
                "condition1",
                "condition1",
            ],
            TIME: [1.0, 1.0, 2.0, 2.0],
            MEASUREMENT: [0.1] * 4,
            NOISE_PARAMETERS: [
                "noiseParOverride1",
                "noiseParOverride1",
                "noiseParOverride2",
                "noiseParOverride2",
            ],
        }
    )

    measurement_df_expected = pd.DataFrame(
        data={
            OBSERVABLE_ID: [
                "obs1__noiseParOverride1__condition1",
                "obs1__noiseParOverride1__condition1",
                "obs1__noiseParOverride2__condition1",
                "obs1__noiseParOverride2__condition1",
            ],
            SIMULATION_CONDITION_ID: [
                "condition1",
                "condition1",
                "condition1",
                "condition1",
            ],
            TIME: [1.0, 1.0, 2.0, 2.0],
            MEASUREMENT: [0.1] * 4,
            NOISE_PARAMETERS: [
                "noiseParOverride1",
                "noiseParOverride1",
                "noiseParOverride2",
                "noiseParOverride2",
            ],
        }
    )

    problem = petab.Problem(
        measurement_df=measurement_df, observable_df=observable_df
    )

    assert petab.lint_problem(problem) is False

    # Ensure having timepoint-specific overrides
    assert (
        petab.lint.measurement_table_has_timepoint_specific_mappings(
            measurement_df
        )
        is True
    )

    petab.flatten_timepoint_specific_output_overrides(problem)

    # Timepoint-specific overrides should be gone now
    assert (
        petab.lint.measurement_table_has_timepoint_specific_mappings(
            problem.measurement_df
        )
        is False
    )

    pd.testing.assert_frame_equal(
        problem.observable_df, observable_df_expected
    )
    pd.testing.assert_frame_equal(
        problem.measurement_df, measurement_df_expected
    )

    assert petab.lint_problem(problem) is False


def test_concat_measurements():
    a = pd.DataFrame({MEASUREMENT: [1.0]})
    b = pd.DataFrame({TIME: [1.0]})

    with tempfile.TemporaryDirectory() as temp_dir:
        filename_a = Path(temp_dir) / "measurements.tsv"
        petab.write_measurement_df(a, filename_a)

        expected = pd.DataFrame({MEASUREMENT: [1.0, nan], TIME: [nan, 1.0]})

        assert expected.equals(
            petab.concat_tables([a, b], petab.measurements.get_measurement_df)
        )

        assert expected.equals(
            petab.concat_tables(
                [filename_a, b], petab.measurements.get_measurement_df
            )
        )


def test_concat_condition_df():
    df1 = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            "par1": [1.1, 1.2],
            "par2": [2.1, 2.2],
            "par3": [3.1, 3.2],
        }
    ).set_index(CONDITION_ID)

    assert df1.equals(petab.concat_tables(df1, petab.get_condition_df))

    df2 = pd.DataFrame(
        data={
            CONDITION_ID: ["condition3"],
            "par1": [1.3],
            "par2": [2.3],
        }
    ).set_index(CONDITION_ID)

    df_expected = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2", "condition3"],
            "par1": [1.1, 1.2, 1.3],
            "par2": [2.1, 2.2, 2.3],
            "par3": [3.1, 3.2, np.nan],
        }
    ).set_index(CONDITION_ID)
    assert df_expected.equals(
        petab.concat_tables((df1, df2), petab.get_condition_df)
    )


def test_get_observable_ids(petab_problem):  # pylint: disable=W0621
    """Test if observable ids functions returns correct value."""
    assert set(petab_problem.get_observable_ids()) == {"obs1"}


def test_parameter_properties(petab_problem):  # pylint: disable=W0621
    """
    Test the petab.Problem functions to get parameter values.
    """
    petab_problem = copy.deepcopy(petab_problem)
    petab_problem.parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["par1", "par2", "par3"],
            LOWER_BOUND: [0, 0.1, 0.1],
            UPPER_BOUND: [100, 100, 200],
            PARAMETER_SCALE: ["lin", "log", "log10"],
            NOMINAL_VALUE: [7, 8, 9],
            ESTIMATE: [1, 1, 0],
        }
    ).set_index(PARAMETER_ID)
    assert petab_problem.x_ids == ["par1", "par2", "par3"]
    assert petab_problem.x_free_ids == ["par1", "par2"]
    assert petab_problem.x_fixed_ids == ["par3"]
    assert petab_problem.lb == [0, 0.1, 0.1]
    assert petab_problem.lb_scaled == [0, np.log(0.1), np.log10(0.1)]
    assert petab_problem.get_lb(fixed=False, scaled=True) == [0, np.log(0.1)]
    assert petab_problem.ub == [100, 100, 200]
    assert petab_problem.ub_scaled == [100, np.log(100), np.log10(200)]
    assert petab_problem.get_ub(fixed=False, scaled=True) == [100, np.log(100)]
    assert petab_problem.x_nominal == [7, 8, 9]
    assert petab_problem.x_nominal_scaled == [7, np.log(8), np.log10(9)]
    assert petab_problem.x_nominal_free == [7, 8]
    assert petab_problem.x_nominal_fixed == [9]
    assert petab_problem.x_nominal_free_scaled == [7, np.log(8)]
    assert petab_problem.x_nominal_fixed_scaled == [np.log10(9)]

    # Check that a missing nominalValues column is handled correctly
    del petab_problem.parameter_df[NOMINAL_VALUE]
    assert len(petab_problem.x_nominal) == 3
    assert np.isnan(petab_problem.x_nominal).all()


def test_to_float_if_float():
    to_float_if_float = petab.core.to_float_if_float

    assert to_float_if_float(1) == 1.0
    assert to_float_if_float("1") == 1.0
    assert to_float_if_float("-1.0") == -1.0
    assert to_float_if_float("1e1") == 10.0
    assert to_float_if_float("abc") == "abc"
    assert to_float_if_float([]) == []


def test_to_files(petab_problem):  # pylint: disable=W0621
    """Test problem.to_files."""
    with tempfile.TemporaryDirectory() as outdir:
        # create target files
        sbml_file = Path(outdir, "model.xml")
        condition_file = Path(outdir, "conditions.tsv")
        measurement_file = Path(outdir, "measurements.tsv")
        parameter_file = Path(outdir, "parameters.tsv")
        observable_file = Path(outdir, "observables.tsv")

        # write contents to files
        petab_problem.to_files(
            model_file=sbml_file,
            condition_file=condition_file,
            measurement_file=measurement_file,
            parameter_file=parameter_file,
            visualization_file=None,
            observable_file=observable_file,
            yaml_file=None,
        )

        # exemplarily load some
        parameter_df = petab.get_parameter_df(parameter_file)
        same_nans = parameter_df.isna() == petab_problem.parameter_df.isna()
        assert (
            ((parameter_df == petab_problem.parameter_df) | same_nans)
            .all()
            .all()
        )


def test_load_remote():
    """Test loading remote files"""

    yaml_url = (
        "https://raw.githubusercontent.com/PEtab-dev/petab_test_suite"
        "/main/petabtests/cases/v1.0.0/sbml/0001/_0001.yaml"
    )
    petab_problem = petab.Problem.from_yaml(yaml_url)

    assert petab_problem.sbml_model is not None
    assert (
        petab_problem.measurement_df is not None
        and not petab_problem.measurement_df.empty
    )


def test_problem_from_yaml_v1_empty():
    """Test loading PEtab version 1 yaml without any files"""
    yaml_config = """
    format_version: 1
    parameter_file:
    problems:
    - condition_files: []
      measurement_files: []
      observable_files: []
      sbml_files: []
    """
    yaml_config = safe_load(StringIO(yaml_config))
    petab.Problem.from_yaml(yaml_config)


def test_problem_from_yaml_v1_multiple_files():
    """Test loading PEtab version 1 yaml with multiple condition / measurement
    / observable files"""
    yaml_config = """
    format_version: 1
    parameter_file:
    problems:
    - condition_files: [conditions1.tsv, conditions2.tsv]
      measurement_files: [measurements1.tsv, measurements2.tsv]
      observable_files: [observables1.tsv, observables2.tsv]
      sbml_files: []
    """

    with tempfile.TemporaryDirectory() as tmpdir:
        yaml_path = Path(tmpdir, "problem.yaml")
        with open(yaml_path, "w") as f:
            f.write(yaml_config)

        for i in (1, 2):
            condition_df = pd.DataFrame(
                {
                    CONDITION_ID: [f"condition{i}"],
                }
            )
            condition_df.set_index([CONDITION_ID], inplace=True)
            petab.write_condition_df(
                condition_df, Path(tmpdir, f"conditions{i}.tsv")
            )

            measurement_df = pd.DataFrame(
                {
                    SIMULATION_CONDITION_ID: [f"condition{i}"],
                    OBSERVABLE_ID: [f"observable{i}"],
                    TIME: [i],
                    MEASUREMENT: [1],
                }
            )
            petab.write_measurement_df(
                measurement_df, Path(tmpdir, f"measurements{i}.tsv")
            )

            observables_df = pd.DataFrame(
                {
                    OBSERVABLE_ID: [f"observable{i}"],
                    OBSERVABLE_FORMULA: [1],
                    NOISE_FORMULA: [1],
                }
            )
            petab.write_observable_df(
                observables_df, Path(tmpdir, f"observables{i}.tsv")
            )

        petab_problem = petab.Problem.from_yaml(yaml_path)

    assert petab_problem.measurement_df.shape[0] == 2
    assert petab_problem.observable_df.shape[0] == 2
    assert petab_problem.condition_df.shape[0] == 2


def test_get_required_parameters_for_parameter_table(petab_problem):
    """Test identification of required parameter table parameters.

    NB: currently, this test only checks that observable parameter placeholders
    in noise formulae are correctly identified as not required in the parameter
    table.
    """
    noise_placeholders = petab.observables.get_output_parameters(
        petab_problem.observable_df,
        petab_problem.model,
        observables=False,
        noise=True,
    )
    # The observable parameter (scaling) appears in the noise formula,
    # as part of the proportional error model.
    assert "observableParameter1_obs1" in noise_placeholders

    required_parameters_for_parameter_table = (
        petab.parameters.get_required_parameters_for_parameter_table(
            model=petab_problem.model,
            condition_df=petab_problem.condition_df,
            observable_df=petab_problem.observable_df,
            measurement_df=petab_problem.measurement_df,
        )
    )
    # The observable parameter is correctly recognized as a placeholder,
    # i.e. does not need to be in the parameter table.
    assert (
        "observableParameter1_obs1"
        not in required_parameters_for_parameter_table
    )



import subprocess
from os import path
from pathlib import Path
from tempfile import TemporaryDirectory

import matplotlib.pyplot as plt
import pytest

import petab
from petab.C import *
from petab.visualize import (
    plot_goodness_of_fit,
    plot_residuals_vs_simulation,
    plot_with_vis_spec,
    plot_without_vis_spec,
)
from petab.visualize.lint import validate_visualization_df
from petab.visualize.plotting import VisSpecParser

# Avoid errors when plotting without X server
plt.switch_backend("agg")

EXAMPLE_DIR = Path(__file__).parents[1] / "doc" / "example"


@pytest.fixture(scope="function")
def close_fig():
    """Close all open matplotlib figures"""
    yield
    plt.close("all")


@pytest.fixture
def data_file_Fujita():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_measurementData.tsv"


@pytest.fixture
def condition_file_Fujita():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_experimentalCondition.tsv"


@pytest.fixture
def data_file_Fujita_wrongNoise():
    return (
        EXAMPLE_DIR
        / "example_Fujita"
        / "Fujita_measurementData_wrongNoise.tsv"
    )


@pytest.fixture
def data_file_Fujita_nanData():
    return (
        EXAMPLE_DIR / "example_Fujita" / "Fujita_measurementData_nanData.tsv"
    )


@pytest.fixture
def simu_file_Fujita():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_simulatedData.tsv"


@pytest.fixture
def simu_file_Fujita_t_inf():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_simulatedData_t_inf.tsv"


@pytest.fixture
def data_file_Fujita_minimal():
    return (
        EXAMPLE_DIR / "example_Fujita" / "Fujita_measurementData_minimal.tsv"
    )


@pytest.fixture
def data_file_Fujita_t_inf():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_measurementData_t_inf.tsv"


@pytest.fixture
def visu_file_Fujita_small():
    return EXAMPLE_DIR / "example_Fujita" / "Fujita_visuSpec_small.tsv"


@pytest.fixture
def visu_file_Fujita_wo_dsid_wo_yvalues():
    return (
        EXAMPLE_DIR / "example_Fujita" / "visuSpecs" / "Fujita_visuSpec_1.tsv"
    )


@pytest.fixture
def visu_file_Fujita_all_obs_with_diff_settings():
    return (
        EXAMPLE_DIR / "example_Fujita" / "visuSpecs" / "Fujita_visuSpec_3.tsv"
    )


@pytest.fixture
def visu_file_Fujita_minimal():
    return (
        EXAMPLE_DIR
        / "example_Fujita"
        / "visuSpecs"
        / "Fujita_visuSpec_mandatory.tsv"
    )


@pytest.fixture
def visu_file_Fujita_empty():
    return (
        EXAMPLE_DIR
        / "example_Fujita"
        / "visuSpecs"
        / "Fujita_visuSpec_empty.tsv"
    )


@pytest.fixture
def visu_file_Fujita_replicates():
    return (
        EXAMPLE_DIR
        / "example_Fujita"
        / "visuSpecs"
        / "Fujita_visuSpec_replicates.tsv"
    )


@pytest.fixture
def data_file_Isensee():
    return EXAMPLE_DIR / "example_Isensee" / "Isensee_measurementData.tsv"


@pytest.fixture
def condition_file_Isensee():
    return (
        EXAMPLE_DIR / "example_Isensee" / "Isensee_experimentalCondition.tsv"
    )


@pytest.fixture
def vis_spec_file_Isensee():
    return (
        EXAMPLE_DIR
        / "example_Isensee"
        / "Isensee_visualizationSpecification.tsv"
    )


@pytest.fixture
def vis_spec_file_Isensee_replicates():
    return (
        EXAMPLE_DIR
        / "example_Isensee"
        / "Isensee_visualizationSpecification_replicates.tsv"
    )


@pytest.fixture
def vis_spec_file_Isensee_scatterplot():
    return (
        EXAMPLE_DIR
        / "example_Isensee"
        / "Isensee_visualizationSpecification_scatterplot.tsv"
    )


@pytest.fixture
def simulation_file_Isensee():
    return EXAMPLE_DIR / "example_Isensee" / "Isensee_simulationData.tsv"


def test_visualization_with_vis_and_sim(
    data_file_Isensee,
    condition_file_Isensee,
    vis_spec_file_Isensee,
    simulation_file_Isensee,
    close_fig,
):
    validate_visualization_df(
        petab.Problem(
            condition_df=petab.get_condition_df(condition_file_Isensee),
            visualization_df=petab.get_visualization_df(vis_spec_file_Isensee),
        )
    )
    plot_with_vis_spec(
        vis_spec_file_Isensee,
        condition_file_Isensee,
        data_file_Isensee,
        simulation_file_Isensee,
    )


def test_visualization_replicates(
    data_file_Isensee,
    condition_file_Isensee,
    vis_spec_file_Isensee_replicates,
    simulation_file_Isensee,
    close_fig,
):
    plot_with_vis_spec(
        vis_spec_file_Isensee_replicates,
        condition_file_Isensee,
        data_file_Isensee,
        simulation_file_Isensee,
    )


def test_visualization_scatterplot(
    data_file_Isensee,
    condition_file_Isensee,
    vis_spec_file_Isensee_scatterplot,
    simulation_file_Isensee,
    close_fig,
):
    plot_with_vis_spec(
        vis_spec_file_Isensee_scatterplot,
        condition_file_Isensee,
        data_file_Isensee,
        simulation_file_Isensee,
    )


def test_visualization_small_visu_file_w_datasetid(
    data_file_Fujita, condition_file_Fujita, visu_file_Fujita_small, close_fig
):
    """
    Test: visualization specification file only with few columns in
    particular datasetId
    (optional columns are optional)
    """
    plot_with_vis_spec(
        visu_file_Fujita_small, condition_file_Fujita, data_file_Fujita
    )


def test_visualization_small_visu_file_wo_datasetid(
    data_file_Fujita,
    condition_file_Fujita,
    visu_file_Fujita_wo_dsid_wo_yvalues,
    close_fig,
):
    """
    Test: visualization specification file only with few columns in
    particular no datasetId column
    (optional columns are optional)
    """
    plot_with_vis_spec(
        visu_file_Fujita_wo_dsid_wo_yvalues,
        condition_file_Fujita,
        data_file_Fujita,
    )


def test_visualization_all_obs_with_diff_settings(
    data_file_Fujita,
    condition_file_Fujita,
    visu_file_Fujita_all_obs_with_diff_settings,
    close_fig,
):
    """
    Test: visualization specification file only with few columns. In
    particular, no datasetId column and no yValues column, but more than one
    plot id. Additionally, having plot id different from 'plot\\d+' for the
    case of vis_spec expansion is tested.
    """
    plot_with_vis_spec(
        visu_file_Fujita_all_obs_with_diff_settings,
        condition_file_Fujita,
        data_file_Fujita,
    )


def test_visualization_minimal_visu_file(
    data_file_Fujita,
    condition_file_Fujita,
    visu_file_Fujita_minimal,
    close_fig,
):
    """
    Test: visualization specification file only with mandatory column plotId
    (optional columns are optional)
    """
    plot_with_vis_spec(
        visu_file_Fujita_minimal, condition_file_Fujita, data_file_Fujita
    )


def test_visualization_empty_visu_file(
    data_file_Fujita, condition_file_Fujita, visu_file_Fujita_empty, close_fig
):
    """
    Test: Empty visualization specification file should default to routine
    for no file at all
    """
    with pytest.warns(UserWarning, match="Visualization table is empty."):
        plot_with_vis_spec(
            visu_file_Fujita_empty, condition_file_Fujita, data_file_Fujita
        )


def test_visualization_minimal_data_file(
    data_file_Fujita_minimal,
    condition_file_Fujita,
    visu_file_Fujita_wo_dsid_wo_yvalues,
    close_fig,
):
    """
    Test visualization, with the case: data file only with mandatory columns
    (optional columns are optional)
    """
    plot_with_vis_spec(
        visu_file_Fujita_wo_dsid_wo_yvalues,
        condition_file_Fujita,
        data_file_Fujita_minimal,
    )


def test_visualization_with_dataset_list(
    data_file_Isensee,
    condition_file_Isensee,
    simulation_file_Isensee,
    close_fig,
):
    datasets = [
        [
            "JI09_150302_Drg345_343_CycNuc__4_ABnOH_and_ctrl",
            "JI09_150302_Drg345_343_CycNuc__4_ABnOH_and_Fsk",
        ],
        [
            "JI09_160201_Drg453-452_CycNuc__ctrl",
            "JI09_160201_Drg453-452_CycNuc__Fsk",
            "JI09_160201_Drg453-452_CycNuc__Sp8_Br_cAMPS_AM",
        ],
    ]

    # TODO: is condition_file needed here
    plot_without_vis_spec(
        condition_file_Isensee, datasets, "dataset", data_file_Isensee
    )

    plot_without_vis_spec(
        condition_file_Isensee,
        datasets,
        "dataset",
        data_file_Isensee,
        simulation_file_Isensee,
    )


def test_visualization_without_datasets(
    data_file_Fujita, condition_file_Fujita, simu_file_Fujita, close_fig
):
    sim_cond_id_list = [
        ["model1_data1"],
        ["model1_data2", "model1_data3"],
        ["model1_data4", "model1_data5"],
        ["model1_data6"],
    ]

    observable_id_list = [["pS6_tot"], ["pEGFR_tot"], ["pAkt_tot"]]

    plot_without_vis_spec(
        condition_file_Fujita,
        sim_cond_id_list,
        "simulation",
        data_file_Fujita,
        plotted_noise=PROVIDED,
    )

    plot_without_vis_spec(
        condition_file_Fujita,
        observable_id_list,
        "observable",
        data_file_Fujita,
        plotted_noise=PROVIDED,
    )

    # with simulations

    plot_without_vis_spec(
        condition_file_Fujita,
        sim_cond_id_list,
        "simulation",
        data_file_Fujita,
        simu_file_Fujita,
        plotted_noise=PROVIDED,
    )

    plot_without_vis_spec(
        condition_file_Fujita,
        observable_id_list,
        "observable",
        data_file_Fujita,
        simu_file_Fujita,
        plotted_noise=PROVIDED,
    )


def test_visualization_only_simulations(
    condition_file_Fujita, simu_file_Fujita, close_fig
):
    sim_cond_id_list = [
        ["model1_data1"],
        ["model1_data2", "model1_data3"],
        ["model1_data4", "model1_data5"],
        ["model1_data6"],
    ]

    observable_id_list = [["pS6_tot"], ["pEGFR_tot"], ["pAkt_tot"]]

    plot_without_vis_spec(
        condition_file_Fujita,
        sim_cond_id_list,
        "simulation",
        simulations_df=simu_file_Fujita,
        plotted_noise=PROVIDED,
    )

    plot_without_vis_spec(
        condition_file_Fujita,
        observable_id_list,
        "observable",
        simulations_df=simu_file_Fujita,
        plotted_noise=PROVIDED,
    )


def test_simple_visualization(
    data_file_Fujita, condition_file_Fujita, close_fig
):
    plot_without_vis_spec(
        condition_file_Fujita, measurements_df=data_file_Fujita
    )
    plot_without_vis_spec(
        condition_file_Fujita,
        measurements_df=data_file_Fujita,
        plotted_noise=PROVIDED,
    )


def test_visualization_with__t_inf(
    data_file_Fujita_t_inf,
    simu_file_Fujita_t_inf,
    condition_file_Fujita,
    visu_file_Fujita_replicates,
    close_fig,
):
    # plot only measurements
    plot_without_vis_spec(
        condition_file_Fujita, measurements_df=data_file_Fujita_t_inf
    )

    # plot only simulation
    plot_without_vis_spec(
        condition_file_Fujita, simulations_df=simu_file_Fujita_t_inf
    )

    # plot both measurements and simulation
    plot_without_vis_spec(
        condition_file_Fujita,
        measurements_df=data_file_Fujita_t_inf,
        simulations_df=simu_file_Fujita_t_inf,
    )

    # plot both measurements and simulation
    plot_with_vis_spec(
        visu_file_Fujita_replicates,
        condition_file_Fujita,
        measurements_df=data_file_Fujita_t_inf,
        simulations_df=simu_file_Fujita_t_inf,
    )


def test_save_plots_to_file(
    data_file_Isensee,
    condition_file_Isensee,
    vis_spec_file_Isensee,
    simulation_file_Isensee,
    close_fig,
):
    with TemporaryDirectory() as temp_dir:
        plot_with_vis_spec(
            vis_spec_file_Isensee,
            condition_file_Isensee,
            data_file_Isensee,
            simulation_file_Isensee,
            subplot_dir=temp_dir,
        )


def test_save_visu_file(data_file_Isensee, condition_file_Isensee):
    with TemporaryDirectory() as temp_dir:
        vis_spec_parser = VisSpecParser(
            condition_file_Isensee, data_file_Isensee
        )
        figure, _ = vis_spec_parser.parse_from_id_list()

        with pytest.warns(
            UserWarning,
            match="please check that datasetId column corresponds to",
        ):
            figure.save_to_tsv(path.join(temp_dir, "visuSpec.tsv"))

        datasets = [
            [
                "JI09_150302_Drg345_343_CycNuc__4_ABnOH_and_ctrl",
                "JI09_150302_Drg345_343_CycNuc__4_ABnOH_and_Fsk",
            ],
            [
                "JI09_160201_Drg453-452_CycNuc__ctrl",
                "JI09_160201_Drg453-452_CycNuc__Fsk",
                "JI09_160201_Drg453-452_CycNuc__Sp8_Br_cAMPS_AM",
            ],
        ]

        vis_spec_parser = VisSpecParser(
            condition_file_Isensee, data_file_Isensee
        )
        figure, _ = vis_spec_parser.parse_from_id_list(
            datasets, group_by="dataset"
        )
        with pytest.warns(
            UserWarning,
            match="please check that datasetId column corresponds to",
        ):
            figure.save_to_tsv(path.join(temp_dir, "visuSpec1.tsv"))


def test_residuals_plot(simu_file_Fujita, close_fig):
    fujita_yaml = EXAMPLE_DIR / "example_Fujita" / "Fujita.yaml"
    fujita_petab_problem = petab.Problem.from_yaml(fujita_yaml)
    plot_residuals_vs_simulation(fujita_petab_problem, simu_file_Fujita)


def test_goodness_of_fit_plot(simu_file_Fujita, close_fig):
    fujita_yaml = EXAMPLE_DIR / "example_Fujita" / "Fujita.yaml"
    fujita_petab_problem = petab.Problem.from_yaml(fujita_yaml)
    plot_goodness_of_fit(fujita_petab_problem, simu_file_Fujita)


def test_cli():
    fujita_dir = EXAMPLE_DIR / "example_Fujita"

    with TemporaryDirectory() as temp_dir:
        args = [
            "petab_visualize",
            "-y",
            str(fujita_dir / "Fujita.yaml"),
            "-s",
            str(fujita_dir / "Fujita_simulatedData.tsv"),
            "-o",
            temp_dir,
        ]
        subprocess.run(args, check=True)


@pytest.mark.filterwarnings("ignore:Visualization table is empty")
@pytest.mark.parametrize(
    "vis_file",
    (
        "vis_spec_file_Isensee",
        "vis_spec_file_Isensee_replicates",
        "vis_spec_file_Isensee_scatterplot",
        "visu_file_Fujita_wo_dsid_wo_yvalues",
        "visu_file_Fujita_all_obs_with_diff_settings",
        "visu_file_Fujita_empty",
        "visu_file_Fujita_minimal",
        "visu_file_Fujita_replicates",
        "visu_file_Fujita_small",
    ),
)
def test_validate(vis_file, request):
    """Check that all test files pass validation."""
    vis_file = request.getfixturevalue(vis_file)
    assert False is validate_visualization_df(
        petab.Problem(visualization_df=petab.get_visualization_df(vis_file))
    )



"""Test COMBINE archive"""
import tempfile
from pathlib import Path

import pandas as pd

import petab
from petab.C import *

# import fixtures
pytest_plugins = [
    "tests.test_petab",
]


def test_combine_archive():
    """Test `create_combine_archive` and `Problem.from_combine`"""

    # Create test files
    import simplesbml

    ss_model = simplesbml.SbmlModel()

    # Create tables with arbitrary content
    measurement_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs1", "obs2"],
            OBSERVABLE_PARAMETERS: ["", "p1;p2"],
            NOISE_PARAMETERS: ["p3;p4", "p5"],
        }
    )

    condition_df = pd.DataFrame(
        data={
            CONDITION_ID: ["condition1", "condition2"],
            CONDITION_NAME: ["", "Condition 2"],
            "fixedParameter1": [1.0, 2.0],
        }
    )
    condition_df.set_index(CONDITION_ID, inplace=True)

    parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["dynamicParameter1", "dynamicParameter2"],
            PARAMETER_NAME: ["", "..."],
        }
    )
    parameter_df.set_index(PARAMETER_ID, inplace=True)

    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["observable_1"],
            OBSERVABLE_FORMULA: ["observable_1"],
            NOISE_FORMULA: [1],
        }
    )
    observable_df.set_index(OBSERVABLE_ID, inplace=True)

    sbml_file_name = "model.xml"
    measurement_file_name = "measurements.tsv"
    condition_file_name = "conditions.tsv"
    parameter_file_name = "parameters.tsv"
    observable_file_name = "observables.tsv"
    yaml_file_name = "test.yaml"

    yaml_config = {
        FORMAT_VERSION: petab.__format_version__,
        PARAMETER_FILE: parameter_file_name,
        PROBLEMS: [
            {
                SBML_FILES: [sbml_file_name],
                MEASUREMENT_FILES: [measurement_file_name],
                CONDITION_FILES: [condition_file_name],
                OBSERVABLE_FILES: [observable_file_name],
            }
        ],
    }

    with tempfile.TemporaryDirectory(
        prefix="petab_test_combine_archive"
    ) as tempdir:
        # Write test data
        outdir = Path(tempdir)
        petab.write_sbml(ss_model.document, outdir / sbml_file_name)
        petab.write_measurement_df(
            measurement_df, outdir / measurement_file_name
        )
        petab.write_parameter_df(parameter_df, outdir / parameter_file_name)
        petab.write_observable_df(observable_df, outdir / observable_file_name)
        petab.write_condition_df(condition_df, outdir / condition_file_name)
        petab.write_yaml(yaml_config, outdir / yaml_file_name)

        archive_file_name = outdir / "test.omex"

        # Create COMBINE archive
        petab.create_combine_archive(
            outdir / yaml_file_name, archive_file_name, family_name="Tester"
        )

        # Read COMBINE archive
        problem = petab.Problem.from_combine(archive_file_name)

        assert problem.parameter_df is not None
        assert problem.condition_df is not None
        assert problem.measurement_df is not None
        assert problem.observable_df is not None



import os
import sys

import pandas as pd
import pytest

sys.path.append(os.getcwd())
import petab  # noqa: E402


def create_test_data():
    # Create test model and data files
    import simplesbml

    ss_model = simplesbml.SbmlModel()
    ss_model.addCompartment(comp_id="compartment_1", vol=1)
    for i in range(1, 4):
        ss_model.addParameter(f"parameter_{i}", i)

    for i in range(1, 5):
        ss_model.addSpecies(f"[species_{i}]", 10 * i)

    ss_model.addAssignmentRule("species_2", "25")

    condition_df = pd.DataFrame(
        {
            petab.CONDITION_ID: ["condition_1"],
            "parameter_3": ["parameter_2"],
            "species_1": [15],
            "species_2": [25],
            "species_3": ["parameter_1"],
            "species_4": ["not_a_model_parameter"],
            "compartment_1": [2],
        }
    )
    condition_df.set_index([petab.CONDITION_ID], inplace=True)

    observable_df = pd.DataFrame(
        {
            petab.OBSERVABLE_ID: ["observable_1"],
            petab.OBSERVABLE_FORMULA: ["2 * species_1"],
        }
    )
    observable_df.set_index([petab.OBSERVABLE_ID], inplace=True)

    measurement_df = pd.DataFrame(
        {
            petab.OBSERVABLE_ID: ["observable_1"],
            petab.SIMULATION_CONDITION_ID: ["condition_1"],
            petab.TIME: [0.0],
        }
    )

    parameter_df = pd.DataFrame(
        {
            petab.PARAMETER_ID: [
                "parameter_1",
                "parameter_2",
                "not_a_model_parameter",
            ],
            petab.PARAMETER_SCALE: [petab.LOG10] * 3,
            petab.NOMINAL_VALUE: [1.25, 2.25, 3.25],
            petab.ESTIMATE: [0, 1, 0],
        }
    )
    parameter_df.set_index([petab.PARAMETER_ID], inplace=True)

    return ss_model, condition_df, observable_df, measurement_df, parameter_df


def check_model(condition_model):
    assert (
        condition_model.getSpecies("species_1").getInitialConcentration() == 15
    )
    assert (
        condition_model.getSpecies("species_2").getInitialConcentration() == 25
    )
    assert (
        condition_model.getSpecies("species_3").getInitialConcentration()
        == 1.25
    )
    assert (
        condition_model.getSpecies("species_4").getInitialConcentration()
        == 3.25
    )
    assert (
        len(condition_model.getListOfInitialAssignments()) == 0
    ), "InitialAssignment not removed"
    assert condition_model.getCompartment("compartment_1").getSize() == 2.0
    assert condition_model.getParameter("parameter_1").getValue() == 1.25
    assert condition_model.getParameter("parameter_2").getValue() == 2.25
    assert condition_model.getParameter("parameter_3").getValue() == 2.25


def test_get_condition_specific_models():
    """Test for petab.sbml.get_condition_specific_models"""
    # retrieve test data
    (
        ss_model,
        condition_df,
        observable_df,
        measurement_df,
        parameter_df,
    ) = create_test_data()

    petab_problem = petab.Problem(
        model=petab.models.sbml_model.SbmlModel(ss_model.model),
        condition_df=condition_df,
        observable_df=observable_df,
        measurement_df=measurement_df,
        parameter_df=parameter_df,
    )

    # create SBML model for condition with parameters updated from problem
    with pytest.warns(
        UserWarning,
        match="An SBML rule was removed to set the "
        "component species_2 to a constant value.",
    ):
        _, condition_model = petab.get_model_for_condition(
            petab_problem, "condition_1"
        )

    check_model(condition_model)



import os
from math import nan

import numpy as np
import pandas as pd

import petab
from petab.C import *
from petab.models.sbml_model import SbmlModel
from petab.parameter_mapping import _apply_parameter_table

# import fixtures
pytest_plugins = [
    "tests.test_petab",
]


class TestGetSimulationToOptimizationParameterMapping(object):
    @staticmethod
    def test_no_condition_specific(condition_df_2_conditions):
        # Trivial case - no condition-specific parameters

        condition_df = condition_df_2_conditions

        measurement_df = pd.DataFrame(
            data={
                OBSERVABLE_ID: ["obs1", "obs2"],
                SIMULATION_CONDITION_ID: ["condition1", "condition2"],
                PREEQUILIBRATION_CONDITION_ID: ["", ""],
                OBSERVABLE_PARAMETERS: ["", ""],
                NOISE_PARAMETERS: ["", ""],
            }
        )

        import simplesbml

        ss_model = simplesbml.SbmlModel()
        ss_model.addParameter("dynamicParameter1", 1.0)
        ss_model.addParameter("dynamicParameter2", 2.0)
        ss_model.addParameter("dynamicParameter3", 3.0)

        # add species, which will have initial concentration in condition table
        #  but which should not show up in mapping
        ss_model.addSpecies("[someSpecies]", 1.0)
        condition_df["someSpecies"] = [0.0, 0.0]

        # Test without parameter table
        expected = [
            (
                {},
                {
                    "dynamicParameter1": 1.0,
                    "dynamicParameter2": 2.0,
                    "dynamicParameter3": 3.0,
                    "fixedParameter1": 1.0,
                },
                {},
                {
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LIN,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
            (
                {},
                {
                    "dynamicParameter1": 1.0,
                    "dynamicParameter2": 2.0,
                    "dynamicParameter3": 3.0,
                    "fixedParameter1": 2.0,
                },
                {},
                {
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LIN,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
        ]

        model = SbmlModel(sbml_model=ss_model.model)
        actual = petab.get_optimization_to_simulation_parameter_mapping(
            model=model,
            measurement_df=measurement_df,
            condition_df=condition_df,
        )
        assert actual == expected

        # Test with parameter table
        parameter_df = pd.DataFrame(
            data={
                PARAMETER_ID: [
                    "dynamicParameter1",
                    "dynamicParameter2",
                    "dynamicParameter3",
                ],
                ESTIMATE: [0, 1, 1],
                NOMINAL_VALUE: [11.0, 12.0, None],
                PARAMETER_SCALE: [LOG, LOG10, LIN],
            }
        )
        parameter_df.set_index(PARAMETER_ID, inplace=True)

        expected = [
            (
                {},
                {
                    "dynamicParameter1": 11.0,
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 1.0,
                },
                {},
                {
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
            (
                {},
                {
                    "dynamicParameter1": 11.0,
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 2.0,
                },
                {},
                {
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            model=model,
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
        )

        assert actual == expected

        # Test with applied scaling

        expected = [
            (
                {},
                {
                    "dynamicParameter1": np.log(11.0),
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 1.0,
                },
                {},
                {
                    "dynamicParameter1": LOG,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
            (
                {},
                {
                    "dynamicParameter1": np.log(11.0),
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 2.0,
                },
                {},
                {
                    "dynamicParameter1": LOG,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            model=model,
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            scaled_parameters=True,
        )

        assert actual == expected

        # Test without fixed overrides

        expected = [
            (
                {},
                {
                    "dynamicParameter1": "dynamicParameter1",
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 1.0,
                },
                {},
                {
                    "dynamicParameter1": LOG,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
            (
                {},
                {
                    "dynamicParameter1": "dynamicParameter1",
                    "dynamicParameter2": "dynamicParameter2",
                    "dynamicParameter3": "dynamicParameter3",
                    "fixedParameter1": 2.0,
                },
                {},
                {
                    "dynamicParameter1": LOG,
                    "dynamicParameter2": LOG10,
                    "dynamicParameter3": LIN,
                    "fixedParameter1": LIN,
                },
            ),
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            model=model,
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            fill_fixed_parameters=False,
        )

        assert actual == expected

    @staticmethod
    def test_all_override(condition_df_2_conditions):
        # Condition-specific parameters overriding original parameters
        condition_df = condition_df_2_conditions

        import simplesbml

        ss_model = simplesbml.SbmlModel()
        ss_model.addParameter("dynamicParameter1", 0.0)
        ss_model.addParameter("dynamicParameter2", 0.0)
        model = SbmlModel(sbml_model=ss_model.model)

        measurement_df = pd.DataFrame(
            data={
                OBSERVABLE_ID: ["obs1", "obs2", "obs1", "obs2"],
                SIMULATION_CONDITION_ID: [
                    "condition1",
                    "condition1",
                    "condition2",
                    "condition2",
                ],
                PREEQUILIBRATION_CONDITION_ID: ["", "", "", ""],
                OBSERVABLE_PARAMETERS: [
                    "obs1par1override;obs1par2cond1override",
                    "obs2par1cond1override",
                    "obs1par1override;obs1par2cond2override",
                    "obs2par1cond2override",
                ],
                NOISE_PARAMETERS: ["", "", "", ""],
            }
        )

        parameter_df = pd.DataFrame(
            data={
                PARAMETER_ID: [
                    "dynamicParameter1",
                    "dynamicParameter2",
                    "obs1par1override",
                    "obs1par2cond1override",
                    "obs1par2cond2override",
                    "obs2par1cond1override",
                    "obs2par1cond2override",
                ],
                ESTIMATE: [1] * 7,
            }
        )
        parameter_df.set_index(PARAMETER_ID, inplace=True)

        expected = [
            (
                {},
                {
                    "fixedParameter1": 1.0,
                    "dynamicParameter1": "dynamicParameter1",
                    "dynamicParameter2": "dynamicParameter2",
                    "observableParameter1_obs1": "obs1par1override",
                    "observableParameter2_obs1": "obs1par2cond1override",
                    "observableParameter1_obs2": "obs2par1cond1override",
                },
                {},
                {
                    "fixedParameter1": LIN,
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LIN,
                    "observableParameter1_obs1": LIN,
                    "observableParameter2_obs1": LIN,
                    "observableParameter1_obs2": LIN,
                },
            ),
            (
                {},
                {
                    "fixedParameter1": 2.0,
                    "dynamicParameter1": "dynamicParameter1",
                    "dynamicParameter2": "dynamicParameter2",
                    "observableParameter1_obs1": "obs1par1override",
                    "observableParameter2_obs1": "obs1par2cond2override",
                    "observableParameter1_obs2": "obs2par1cond2override",
                },
                {},
                {
                    "fixedParameter1": LIN,
                    "dynamicParameter1": LIN,
                    "dynamicParameter2": LIN,
                    "observableParameter1_obs1": LIN,
                    "observableParameter2_obs1": LIN,
                    "observableParameter1_obs2": LIN,
                },
            ),
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            model=model,
            parameter_df=parameter_df,
        )
        assert actual == expected

        # For one case we test parallel execution, which must yield the same
        # result
        os.environ[petab.ENV_NUM_THREADS] = "4"
        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            model=model,
            parameter_df=parameter_df,
        )
        assert actual == expected

    @staticmethod
    def test_partial_override(condition_df_2_conditions):
        # Condition-specific parameters, keeping original parameters
        condition_df = pd.DataFrame(
            data={
                "conditionId": ["condition1", "condition2"],
                "conditionName": ["", "Condition 2"],
                "fixedParameter1": [1.0, 2.0],
                "fixedParameter2": [nan, 2.5],
            }
        )
        condition_df.set_index("conditionId", inplace=True)

        import simplesbml

        ss_model = simplesbml.SbmlModel()
        ss_model.addParameter("fixedParameter1", 0.5)
        ss_model.addParameter("fixedParameter2", 1.0)
        ss_model.addParameter("dynamicParameter1", 0.0)
        ss_model.addParameter("observableParameter1_obs1", 0.0)
        ss_model.addParameter("observableParameter2_obs1", 0.0)
        ss_model.addParameter("observableParameter1_obs2", 0.0)

        measurement_df = pd.DataFrame(
            data={
                OBSERVABLE_ID: ["obs1", "obs2", "obs1", "obs2"],
                SIMULATION_CONDITION_ID: [
                    "condition1",
                    "condition1",
                    "condition2",
                    "condition2",
                ],
                PREEQUILIBRATION_CONDITION_ID: ["", "", "", ""],
                OBSERVABLE_PARAMETERS: [
                    "obs1par1override;obs1par2cond1override",
                    "",
                    "obs1par1override;obs1par2cond2override",
                    "obs2par1cond2override",
                ],
                NOISE_PARAMETERS: ["", "", "", ""],
            }
        )

        parameter_df = pd.DataFrame(
            data={
                PARAMETER_ID: [
                    "dynamicParameter1",
                    "obs1par1override",
                    "obs1par2cond1override",
                    "obs1par2cond2override",
                    "obs2par1cond2override",
                ],
                ESTIMATE: [1, 1, 1, 1, 1],
            }
        )
        parameter_df.set_index(PARAMETER_ID, inplace=True)

        expected = [
            (
                {},
                {
                    "fixedParameter1": 1.0,
                    "fixedParameter2": 1.0,
                    "dynamicParameter1": "dynamicParameter1",
                    "observableParameter1_obs1": "obs1par1override",
                    "observableParameter2_obs1": "obs1par2cond1override",
                    "observableParameter1_obs2": np.nan,
                },
                {},
                {
                    "fixedParameter1": LIN,
                    "fixedParameter2": LIN,
                    "dynamicParameter1": LIN,
                    "observableParameter1_obs1": LIN,
                    "observableParameter2_obs1": LIN,
                    "observableParameter1_obs2": LIN,
                },
            ),
            (
                {},
                {
                    "fixedParameter1": 2.0,
                    "fixedParameter2": 2.5,
                    "dynamicParameter1": "dynamicParameter1",
                    "observableParameter1_obs1": "obs1par1override",
                    "observableParameter2_obs1": "obs1par2cond2override",
                    "observableParameter1_obs2": "obs2par1cond2override",
                },
                {},
                {
                    "fixedParameter1": LIN,
                    "fixedParameter2": LIN,
                    "dynamicParameter1": LIN,
                    "observableParameter1_obs1": LIN,
                    "observableParameter2_obs1": LIN,
                    "observableParameter1_obs2": LIN,
                },
            ),
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            model=petab.models.sbml_model.SbmlModel(ss_model.model),
            parameter_df=parameter_df,
        )

        # Comparison with NaN containing expected results fails after pickling!
        # Need to test first for correct NaNs, then for the rest.
        assert np.isnan(expected[0][1]["observableParameter1_obs2"])
        assert np.isnan(actual[0][1]["observableParameter1_obs2"])
        expected[0][1]["observableParameter1_obs2"] = 0.0
        actual[0][1]["observableParameter1_obs2"] = 0.0

        assert actual == expected

    @staticmethod
    def test_parameterized_condition_table():
        condition_df = pd.DataFrame(
            data={
                CONDITION_ID: ["condition1", "condition2", "condition3"],
                CONDITION_NAME: ["", "Condition 2", ""],
                "dynamicParameter1": [
                    "dynamicOverride1_1",
                    "dynamicOverride1_2",
                    0,
                ],
            }
        )
        condition_df.set_index(CONDITION_ID, inplace=True)

        measurement_df = pd.DataFrame(
            data={
                SIMULATION_CONDITION_ID: [
                    "condition1",
                    "condition2",
                    "condition3",
                ],
                OBSERVABLE_ID: ["obs1", "obs2", "obs1"],
                OBSERVABLE_PARAMETERS: "",
                NOISE_PARAMETERS: "",
            }
        )

        parameter_df = pd.DataFrame(
            data={
                PARAMETER_ID: ["dynamicOverride1_1", "dynamicOverride1_2"],
                PARAMETER_NAME: ["", "..."],
                ESTIMATE: [1, 1],
            }
        )
        parameter_df.set_index(PARAMETER_ID, inplace=True)

        import simplesbml

        ss_model = simplesbml.SbmlModel()
        ss_model.addParameter("dynamicParameter1", 1.0)

        assert petab.get_model_parameters(ss_model.model) == [
            "dynamicParameter1"
        ]

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            model=petab.models.sbml_model.SbmlModel(ss_model.model),
        )

        expected = [
            (
                {},
                {"dynamicParameter1": "dynamicOverride1_1"},
                {},
                {"dynamicParameter1": LIN},
            ),
            (
                {},
                {"dynamicParameter1": "dynamicOverride1_2"},
                {},
                {"dynamicParameter1": LIN},
            ),
            ({}, {"dynamicParameter1": 0}, {}, {"dynamicParameter1": LIN}),
        ]

        assert actual == expected

    @staticmethod
    def test_parameterized_condition_table_changed_scale():
        """Test overriding a dynamic parameter `overridee` with
        - a log10 parameter to be estimated (condition 1)
        - lin parameter not estimated (condition2)
        - log10 parameter not estimated (condition 3)
        - constant override (condition 4)"""

        # overridden parameter
        overridee_id = "overridee"

        # set up model
        import simplesbml

        ss_model = simplesbml.SbmlModel()
        ss_model.addParameter(overridee_id, 2.0)
        assert petab.get_model_parameters(ss_model.model) == [overridee_id]
        assert petab.get_model_parameters(
            ss_model.model, with_values=True
        ) == {overridee_id: 2.0}

        # set up condition table
        condition_df = pd.DataFrame(
            data={
                CONDITION_ID: [
                    "condition1",
                    "condition2",
                    "condition3",
                    "condition4",
                ],
                overridee_id: [
                    "dynamicOverrideLog10",
                    "fixedOverrideLin",
                    "fixedOverrideLog10",
                    10.0,
                ],
            }
        )
        condition_df.set_index("conditionId", inplace=True)

        # set up measurement table
        measurement_df = pd.DataFrame(
            data={
                SIMULATION_CONDITION_ID: [
                    "condition1",
                    "condition2",
                    "condition3",
                    "condition4",
                ],
                OBSERVABLE_ID: ["obs1", "obs2", "obs1", "obs2"],
                OBSERVABLE_PARAMETERS: "",
                NOISE_PARAMETERS: "",
            }
        )

        # set up parameter table
        parameter_df = pd.DataFrame(
            data={
                PARAMETER_ID: [
                    "dynamicOverrideLog10",
                    "fixedOverrideLin",
                    "fixedOverrideLog10",
                ],
                ESTIMATE: [1, 0, 0],
                NOMINAL_VALUE: [np.nan, -2, 1000],
                PARAMETER_SCALE: [LOG10, LIN, LOG10],
            }
        )
        parameter_df.set_index(PARAMETER_ID, inplace=True)

        # test without preequilibration condition; unscaled known parameters

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            model=petab.models.sbml_model.SbmlModel(ss_model.model),
        )

        expected = [
            (
                {},
                {overridee_id: "dynamicOverrideLog10"},
                {},
                {overridee_id: LOG10},
            ),
            ({}, {overridee_id: -2.0}, {}, {overridee_id: LIN}),
            # not scaled:
            ({}, {overridee_id: 1000.0}, {}, {overridee_id: LIN}),
            ({}, {overridee_id: 10.0}, {}, {overridee_id: LIN}),
        ]

        assert actual == expected

        # test without preequilibration condition; scaled known parameters

        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            model=petab.models.sbml_model.SbmlModel(ss_model.model),
            scaled_parameters=True,
        )

        expected = [
            (
                {},
                {overridee_id: "dynamicOverrideLog10"},
                {},
                {overridee_id: LOG10},
            ),
            ({}, {overridee_id: -2.0}, {}, {overridee_id: LIN}),
            # scaled fixedOverrideLog10:
            ({}, {overridee_id: 3.0}, {}, {overridee_id: LOG10}),
            ({}, {overridee_id: 10.0}, {}, {overridee_id: LIN}),
        ]

        assert actual == expected

        # Add preeq condition

        measurement_df[PREEQUILIBRATION_CONDITION_ID] = [
            "condition1",
            "condition1",
            "condition3",
            "condition3",
        ]
        actual = petab.get_optimization_to_simulation_parameter_mapping(
            measurement_df=measurement_df,
            condition_df=condition_df,
            parameter_df=parameter_df,
            model=petab.models.sbml_model.SbmlModel(ss_model.model),
        )

        expected = [
            (
                {overridee_id: "dynamicOverrideLog10"},
                {overridee_id: "dynamicOverrideLog10"},
                {overridee_id: LOG10},
                {overridee_id: LOG10},
            ),
            (
                {overridee_id: "dynamicOverrideLog10"},
                {overridee_id: -2.0},
                {overridee_id: LOG10},
                {overridee_id: LIN},
            ),
            # not rescaled:
            (
                {overridee_id: 1000.0},
                {overridee_id: 1000.0},
                {overridee_id: LIN},
                {overridee_id: LIN},
            ),
            (
                {overridee_id: 1000.0},
                {overridee_id: 10.0},
                {overridee_id: LIN},
                {overridee_id: LIN},
            ),
        ]

        assert actual == expected


def test_fill_in_nominal_values():
    parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["estimated", "not_estimated"],
            NOMINAL_VALUE: [0.0, 2.0],
            ESTIMATE: [1, 0],
        }
    )
    parameter_df.set_index([PARAMETER_ID], inplace=True)

    par_mapping = {"estimated": "estimated", "not_estimated": "not_estimated"}
    scale_mapping = {"estimated": LIN, "not_estimated": LIN}
    _apply_parameter_table(par_mapping, scale_mapping, parameter_df)

    expected_par = {"estimated": "estimated", "not_estimated": 2.0}
    expected_scale = {"estimated": LIN, "not_estimated": LIN}

    assert expected_par == par_mapping
    assert expected_scale == scale_mapping



# pylint: disable:invalid-name
"""
This file contains constant definitions.
"""

import math as _math

# MEASUREMENTS

#:
OBSERVABLE_ID = "observableId"

#:
PREEQUILIBRATION_CONDITION_ID = "preequilibrationConditionId"

#:
SIMULATION_CONDITION_ID = "simulationConditionId"

#:
MEASUREMENT = "measurement"

#:
TIME = "time"

#: Time value that indicates steady-state measurements
TIME_STEADY_STATE = _math.inf

#:
OBSERVABLE_PARAMETERS = "observableParameters"

#:
NOISE_PARAMETERS = "noiseParameters"

#:
DATASET_ID = "datasetId"

#:
REPLICATE_ID = "replicateId"

#: Mandatory columns of measurement table
MEASUREMENT_DF_REQUIRED_COLS = [
    OBSERVABLE_ID,
    SIMULATION_CONDITION_ID,
    MEASUREMENT,
    TIME,
]

#: Optional columns of measurement table
MEASUREMENT_DF_OPTIONAL_COLS = [
    PREEQUILIBRATION_CONDITION_ID,
    OBSERVABLE_PARAMETERS,
    NOISE_PARAMETERS,
    DATASET_ID,
    REPLICATE_ID,
]

#: Measurement table columns
MEASUREMENT_DF_COLS = [
    MEASUREMENT_DF_REQUIRED_COLS[0],
    MEASUREMENT_DF_OPTIONAL_COLS[0],
    *MEASUREMENT_DF_REQUIRED_COLS[1:],
    *MEASUREMENT_DF_OPTIONAL_COLS[1:],
]


# PARAMETERS

#:
PARAMETER_ID = "parameterId"
#:
PARAMETER_NAME = "parameterName"
#:
PARAMETER_SCALE = "parameterScale"
#:
LOWER_BOUND = "lowerBound"
#:
UPPER_BOUND = "upperBound"
#:
NOMINAL_VALUE = "nominalValue"
#:
ESTIMATE = "estimate"
#:
INITIALIZATION_PRIOR_TYPE = "initializationPriorType"
#:
INITIALIZATION_PRIOR_PARAMETERS = "initializationPriorParameters"
#:
OBJECTIVE_PRIOR_TYPE = "objectivePriorType"
#:
OBJECTIVE_PRIOR_PARAMETERS = "objectivePriorParameters"

#: Mandatory columns of parameter table
PARAMETER_DF_REQUIRED_COLS = [
    PARAMETER_ID,
    PARAMETER_SCALE,
    LOWER_BOUND,
    UPPER_BOUND,
    ESTIMATE,
]

#: Optional columns of parameter table
PARAMETER_DF_OPTIONAL_COLS = [
    PARAMETER_NAME,
    NOMINAL_VALUE,
    INITIALIZATION_PRIOR_TYPE,
    INITIALIZATION_PRIOR_PARAMETERS,
    OBJECTIVE_PRIOR_TYPE,
    OBJECTIVE_PRIOR_PARAMETERS,
]

#: Parameter table columns
PARAMETER_DF_COLS = [
    PARAMETER_DF_REQUIRED_COLS[0],
    PARAMETER_DF_OPTIONAL_COLS[0],
    *PARAMETER_DF_REQUIRED_COLS[1:],
    *PARAMETER_DF_OPTIONAL_COLS[1:],
]

#:
INITIALIZATION = "initialization"
#:
OBJECTIVE = "objective"


# CONDITIONS

#:
CONDITION_ID = "conditionId"
#:
CONDITION_NAME = "conditionName"


# OBSERVABLES

#:
OBSERVABLE_NAME = "observableName"
#:
OBSERVABLE_FORMULA = "observableFormula"
#:
NOISE_FORMULA = "noiseFormula"
#:
OBSERVABLE_TRANSFORMATION = "observableTransformation"
#:
NOISE_DISTRIBUTION = "noiseDistribution"

#: Mandatory columns of observables table
OBSERVABLE_DF_REQUIRED_COLS = [
    OBSERVABLE_ID,
    OBSERVABLE_FORMULA,
    NOISE_FORMULA,
]

#: Optional columns of observables table
OBSERVABLE_DF_OPTIONAL_COLS = [
    OBSERVABLE_NAME,
    OBSERVABLE_TRANSFORMATION,
    NOISE_DISTRIBUTION,
]

#: Observables table columns
OBSERVABLE_DF_COLS = [
    *OBSERVABLE_DF_REQUIRED_COLS,
    *OBSERVABLE_DF_OPTIONAL_COLS,
]


# TRANSFORMATIONS

#:
LIN = "lin"
#:
LOG = "log"
#:
LOG10 = "log10"
#: Supported observable transformations
OBSERVABLE_TRANSFORMATIONS = [LIN, LOG, LOG10]


# NOISE MODELS

#:
UNIFORM = "uniform"
#:
PARAMETER_SCALE_UNIFORM = "parameterScaleUniform"
#:
NORMAL = "normal"
#:
PARAMETER_SCALE_NORMAL = "parameterScaleNormal"
#:
LAPLACE = "laplace"
#:
PARAMETER_SCALE_LAPLACE = "parameterScaleLaplace"
#:
LOG_NORMAL = "logNormal"
#:
LOG_LAPLACE = "logLaplace"

#: Supported prior types
PRIOR_TYPES = [
    UNIFORM,
    NORMAL,
    LAPLACE,
    LOG_NORMAL,
    LOG_LAPLACE,
    PARAMETER_SCALE_UNIFORM,
    PARAMETER_SCALE_NORMAL,
    PARAMETER_SCALE_LAPLACE,
]

#: Supported noise distributions
NOISE_MODELS = [NORMAL, LAPLACE]


# VISUALIZATION

#:
PLOT_ID = "plotId"
#:
PLOT_NAME = "plotName"
#:
PLOT_TYPE_SIMULATION = "plotTypeSimulation"
#:
PLOT_TYPE_DATA = "plotTypeData"
#:
X_VALUES = "xValues"
#:
X_OFFSET = "xOffset"
#:
X_LABEL = "xLabel"
#:
X_SCALE = "xScale"
#:
Y_VALUES = "yValues"
#:
Y_OFFSET = "yOffset"
#:
Y_LABEL = "yLabel"
#:
Y_SCALE = "yScale"
#:
LEGEND_ENTRY = "legendEntry"

#: Mandatory columns of visualization table
VISUALIZATION_DF_REQUIRED_COLS = [PLOT_ID]

#: Optional columns of visualization table
VISUALIZATION_DF_OPTIONAL_COLS = [
    PLOT_NAME,
    PLOT_TYPE_SIMULATION,
    PLOT_TYPE_DATA,
    X_VALUES,
    X_OFFSET,
    X_LABEL,
    X_SCALE,
    Y_VALUES,
    Y_OFFSET,
    Y_LABEL,
    Y_SCALE,
    LEGEND_ENTRY,
    DATASET_ID,
]

#: Visualization table columns
VISUALIZATION_DF_COLS = [
    *VISUALIZATION_DF_REQUIRED_COLS,
    *VISUALIZATION_DF_OPTIONAL_COLS,
]

#: Visualization table columns that contain subplot specifications
VISUALIZATION_DF_SUBPLOT_LEVEL_COLS = [
    PLOT_ID,
    PLOT_NAME,
    PLOT_TYPE_SIMULATION,
    PLOT_TYPE_DATA,
    X_LABEL,
    X_SCALE,
    Y_LABEL,
    Y_SCALE,
]

#: Visualization table columns that contain single plot specifications
VISUALIZATION_DF_SINGLE_PLOT_LEVEL_COLS = [
    X_VALUES,
    X_OFFSET,
    Y_VALUES,
    Y_OFFSET,
    LEGEND_ENTRY,
    DATASET_ID,
]

#:
LINE_PLOT = "LinePlot"
#:
BAR_PLOT = "BarPlot"
#:
SCATTER_PLOT = "ScatterPlot"
#: Supported plot types
PLOT_TYPES_SIMULATION = [LINE_PLOT, BAR_PLOT, SCATTER_PLOT]

#: Supported xScales
X_SCALES = [LIN, LOG, LOG10]

#: Supported yScales
Y_SCALES = [LIN, LOG, LOG10]


#:
MEAN_AND_SD = "MeanAndSD"
#:
MEAN_AND_SEM = "MeanAndSEM"
#:
REPLICATE = "replicate"
#:
PROVIDED = "provided"
#: Supported settings for handling replicates
PLOT_TYPES_DATA = [MEAN_AND_SD, MEAN_AND_SEM, REPLICATE, PROVIDED]


# YAML
#:
FORMAT_VERSION = "format_version"
#:
PARAMETER_FILE = "parameter_file"
#:
PROBLEMS = "problems"
#:
SBML_FILES = "sbml_files"
#:
MODEL_FILES = "model_files"
#:
MODEL_LOCATION = "location"
#:
MODEL_LANGUAGE = "language"
#:
CONDITION_FILES = "condition_files"
#:
MEASUREMENT_FILES = "measurement_files"
#:
OBSERVABLE_FILES = "observable_files"
#:
VISUALIZATION_FILES = "visualization_files"
#:
MAPPING_FILES = "mapping_files"
#:
EXTENSIONS = "extensions"


# MAPPING
#:
PETAB_ENTITY_ID = "petabEntityId"
#:
MODEL_ENTITY_ID = "modelEntityId"
#:
MAPPING_DF_REQUIRED_COLS = [PETAB_ENTITY_ID, MODEL_ENTITY_ID]

# MORE

#:
SIMULATION = "simulation"
#:
RESIDUAL = "residual"
#:
NOISE_VALUE = "noiseValue"

# separator for multiple parameter values (bounds, observableParameters, ...)
PARAMETER_SEPARATOR = ";"



"""Tests for petab/parameters.py"""
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import petab
from petab.C import *


def test_get_optimization_parameter_scaling():
    """Test get_optimization_parameter_scaling"""
    df = pd.DataFrame(
        data={
            PARAMETER_ID: ["p1", "p2", "p3"],
            ESTIMATE: [1, 0, 1],
            PARAMETER_SCALE: [LIN, LOG, LOG10],
        }
    )
    df.set_index(PARAMETER_ID, inplace=True)

    # parameter and scale
    expected = dict(p1=LIN, p3=LOG10)

    actual = petab.get_optimization_parameter_scaling(df)

    assert actual == expected


def test_get_optimization_parameters():
    """Test get_optimization_parameters"""
    df = pd.DataFrame(
        data={
            PARAMETER_ID: ["p1", "p2", "p3"],
            ESTIMATE: [1, 0, 1],
        }
    )
    df.set_index(PARAMETER_ID, inplace=True)

    expected = ["p1", "p3"]

    actual = petab.get_optimization_parameters(df)

    assert actual == expected


def test_get_parameter_df():
    """Test parameters.get_parameter_df."""
    # parameter df missing ids
    parameter_df = pd.DataFrame(
        data={
            PARAMETER_NAME: ["parname1", "parname2"],
        }
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        parameter_df.to_csv(fh, sep="\t", index=False)

    with pytest.raises(KeyError):
        petab.get_parameter_df(file_name)

    # with ids
    parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["par1", "par2"],
            PARAMETER_NAME: ["parname1", "parname2"],
        }
    )
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        parameter_df.to_csv(fh, sep="\t", index=False)

    df = petab.get_parameter_df(file_name)
    assert (df == parameter_df.set_index(PARAMETER_ID)).all().all()

    # Test parameter subset files
    with tempfile.TemporaryDirectory() as directory:
        parameter_dfs, parameter_files = ({}, {})
        parameter_dfs["complete"] = pd.DataFrame(
            data={
                PARAMETER_ID: ["id1", "id2", "id3"],
                PARAMETER_NAME: ["name1", "name2", "name3"],
            }
        )
        parameter_dfs["subset1"] = pd.DataFrame(
            data={
                PARAMETER_ID: ["id1", "id2"],
                PARAMETER_NAME: ["name1", "name2"],
            }
        )
        parameter_dfs["subset2_strict"] = pd.DataFrame(
            data={PARAMETER_ID: ["id3"], PARAMETER_NAME: ["name3"]}
        )
        parameter_dfs["subset2_redundance"] = pd.DataFrame(
            data={
                PARAMETER_ID: ["id2", "id3"],
                PARAMETER_NAME: ["name2", "name3"],
            }
        )
        parameter_dfs["subset2_contradiction"] = pd.DataFrame(
            data={
                PARAMETER_ID: ["id2", "id3"],
                PARAMETER_NAME: ["different_name2", "name3"],
            }
        )
        for name, df in parameter_dfs.items():
            with tempfile.NamedTemporaryFile(
                mode="w", delete=False, dir=directory
            ) as fh:
                parameter_files[name] = fh.name
                parameter_dfs[name].to_csv(fh, sep="\t", index=False)
        # Check that subset files are correctly combined
        assert petab.get_parameter_df(parameter_files["complete"]).equals(
            petab.get_parameter_df(
                [parameter_files["subset1"], parameter_files["subset2_strict"]]
            )
        )
        # Ensure an error is raised if there exist parameterId duplicates
        # with identical parameter definitions
        with pytest.raises(ValueError):
            petab.get_parameter_df(
                [
                    parameter_files["subset1"],
                    parameter_files["subset2_redundance"],
                ]
            )
        # with non-identical parameter definitions
        with pytest.raises(ValueError):
            petab.get_parameter_df(
                [
                    parameter_files["subset1"],
                    parameter_files["subset2_contradiction"],
                ]
            )

    # Ensure that parameters that differ only by parameterId
    # are recognized as distinct
    with tempfile.TemporaryDirectory() as directory:
        parameter_dfs, parameter_files = ({}, {})
        parameter_dfs["complete"] = pd.DataFrame(
            data={
                PARAMETER_ID: ["id1", "id2", "id3", "id4"],
                NOMINAL_VALUE: [1, 1, 1, 1],
            }
        )
        parameter_dfs["subset1"] = pd.DataFrame(
            data={PARAMETER_ID: ["id1", "id2"], NOMINAL_VALUE: [1, 1]}
        )
        parameter_dfs["subset2"] = pd.DataFrame(
            data={PARAMETER_ID: ["id3", "id4"], NOMINAL_VALUE: [1, 1]}
        )
        for name, df in parameter_dfs.items():
            with tempfile.NamedTemporaryFile(
                mode="w", delete=False, dir=directory
            ) as fh:
                parameter_files[name] = fh.name
                parameter_dfs[name].to_csv(fh, sep="\t", index=False)
        # from one parameter file
        df_template = parameter_dfs["complete"].set_index(PARAMETER_ID)
        df_test = petab.get_parameter_df(parameter_files["complete"])
        assert (df_template == df_test).all().all()
        # several parameter files
        assert petab.get_parameter_df(parameter_files["complete"]).equals(
            petab.get_parameter_df(
                [parameter_files["subset1"], parameter_files["subset2"]]
            )
        )


def test_write_parameter_df():
    """Test parameters.write_parameter_df."""
    parameter_df = pd.DataFrame(
        data={
            PARAMETER_ID: ["par1", "par2"],
            # Test utf8 characters
            PARAMETER_NAME: ["ɑ", "β"],
        }
    ).set_index(PARAMETER_ID)

    with tempfile.TemporaryDirectory() as temp_dir:
        file_name = Path(temp_dir) / "parameters.tsv"
        petab.write_parameter_df(parameter_df, file_name)
        re_df = petab.get_parameter_df(file_name)
        assert (parameter_df == re_df).all().all()


def test_normalize_parameter_df():
    """Check parameters.normalize_parameter_df."""
    parameter_df = pd.DataFrame(
        {
            PARAMETER_ID: ["par0", "par1", "par2"],
            PARAMETER_SCALE: [LOG10, LOG10, LIN],
            NOMINAL_VALUE: [1e-2, 1e-3, 1e-4],
            ESTIMATE: [1, 1, 0],
            LOWER_BOUND: [1e-5, 1e-6, 1e-7],
            UPPER_BOUND: [1e5, 1e6, 1e7],
        }
    ).set_index(PARAMETER_ID)

    actual = petab.normalize_parameter_df(parameter_df)

    expected = parameter_df.copy(deep=True)
    expected[PARAMETER_NAME] = parameter_df.reset_index()[PARAMETER_ID]
    expected[INITIALIZATION_PRIOR_TYPE] = [PARAMETER_SCALE_UNIFORM] * 3
    expected[INITIALIZATION_PRIOR_PARAMETERS] = ["-5;5", "-6;6", "1e-7;1e7"]
    expected[OBJECTIVE_PRIOR_TYPE] = [PARAMETER_SCALE_UNIFORM] * 3
    expected[OBJECTIVE_PRIOR_PARAMETERS] = ["-5;5", "-6;6", "1e-7;1e7"]

    # check ids
    assert list(actual.index.values) == list(expected.index.values)

    # check if basic columns match
    for col in PARAMETER_DF_COLS[1:]:
        if col in [
            INITIALIZATION_PRIOR_PARAMETERS,
            OBJECTIVE_PRIOR_PARAMETERS,
        ]:
            continue
        assert (
            (actual[col] == expected[col])
            | (actual[col].isnull() == expected[col].isnull())
        ).all()

    # check if prior parameters match
    for col in [INITIALIZATION_PRIOR_PARAMETERS, OBJECTIVE_PRIOR_PARAMETERS]:
        for (_, actual_row), (_, expected_row) in zip(
            actual.iterrows(), expected.iterrows()
        ):
            actual_pars = tuple(
                [float(val) for val in actual_row[col].split(";")]
            )
            expected_pars = tuple(
                [float(val) for val in expected_row[col].split(";")]
            )

            assert actual_pars == expected_pars

    # check is a projection
    actual2 = petab.normalize_parameter_df(actual)
    assert (
        ((actual == actual2) | (actual.isnull() == actual2.isnull()))
        .all()
        .all()
    )

    # check is valid petab
    petab.check_parameter_df(actual)


def test_scale_unscale():
    """Test the parameter scaling functions."""
    par = 2.5
    # scale
    assert petab.scale(par, LIN) == par
    assert petab.scale(par, LOG) == np.log(par)
    assert petab.scale(par, LOG10) == np.log10(par)
    # unscale
    assert petab.unscale(par, LIN) == par
    assert petab.unscale(par, LOG) == np.exp(par)
    assert petab.unscale(par, LOG10) == 10**par

    # map scale
    assert list(petab.map_scale([par] * 3, [LIN, LOG, LOG10])) == [
        par,
        np.log(par),
        np.log10(par),
    ]
    # map unscale
    assert list(petab.map_unscale([par] * 3, [LIN, LOG, LOG10])) == [
        par,
        np.exp(par),
        10**par,
    ]

    # map broadcast
    assert list(petab.map_scale([par, 2 * par], LOG)) == list(
        np.log([par, 2 * par])
    )
    assert list(petab.map_unscale([par, 2 * par], LOG)) == list(
        np.exp([par, 2 * par])
    )



"""Tests for petab.observables"""
import tempfile
from pathlib import Path

import pandas as pd
import pytest

import petab
from petab.C import *

# import fixtures
pytest_plugins = [
    "tests.test_petab",
]


def test_get_observable_df():
    """Test measurements.get_measurement_df."""
    # without id
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_NAME: ["observable name 1"],
            OBSERVABLE_FORMULA: ["observable_1"],
            NOISE_FORMULA: [1],
        }
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        observable_df.to_csv(fh, sep="\t", index=False)

    with pytest.raises(KeyError):
        petab.get_observable_df(file_name)

    # with id
    observable_df[OBSERVABLE_ID] = ["observable_1"]

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        file_name = fh.name
        observable_df.to_csv(fh, sep="\t", index=False)

    df = petab.get_observable_df(file_name)
    assert (df == observable_df.set_index(OBSERVABLE_ID)).all().all()

    # test other arguments
    assert (
        (petab.get_observable_df(observable_df) == observable_df).all().all()
    )
    assert petab.get_observable_df(None) is None


def test_write_observable_df():
    """Test measurements.get_measurement_df."""
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["observable_1"],
            OBSERVABLE_NAME: ["observable name 1"],
            OBSERVABLE_FORMULA: ["observable_1"],
            NOISE_FORMULA: [1],
        }
    ).set_index(OBSERVABLE_ID)

    with tempfile.TemporaryDirectory() as temp_dir:
        file_name = Path(temp_dir) / "observables.tsv"
        petab.write_observable_df(observable_df, file_name)
        re_df = petab.get_observable_df(file_name)
        assert (observable_df == re_df).all().all()


def test_get_output_parameters():
    """Test measurements.get_output_parameters."""
    # sbml model
    import simplesbml

    from petab.models.sbml_model import SbmlModel

    ss_model = simplesbml.SbmlModel()
    ss_model.addParameter("fixedParameter1", 1.0)
    ss_model.addParameter("observable_1", 1.0)

    # observable file
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["observable_1"],
            OBSERVABLE_NAME: ["observable name 1"],
            OBSERVABLE_FORMULA: ["observable_1 * scaling + offset"],
            NOISE_FORMULA: [1],
        }
    ).set_index(OBSERVABLE_ID)

    output_parameters = petab.get_output_parameters(
        observable_df, SbmlModel(sbml_model=ss_model.model)
    )

    assert output_parameters == ["offset", "scaling"]

    # test sympy-special symbols (e.g. N, beta, ...)
    # see https://github.com/ICB-DCM/pyPESTO/issues/1048
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["observable_1"],
            OBSERVABLE_NAME: ["observable name 1"],
            OBSERVABLE_FORMULA: ["observable_1 * N + beta"],
            NOISE_FORMULA: [1],
        }
    ).set_index(OBSERVABLE_ID)

    output_parameters = petab.get_output_parameters(
        observable_df, SbmlModel(sbml_model=ss_model.model)
    )

    assert output_parameters == ["N", "beta"]


def test_get_formula_placeholders():
    """Test get_formula_placeholders"""

    # no placeholder
    assert petab.get_formula_placeholders("1.0", "any", "observable") == []

    # multiple placeholders
    assert petab.get_formula_placeholders(
        "observableParameter1_twoParams * "
        "observableParameter2_twoParams + otherParam",
        "twoParams",
        "observable",
    ) == ["observableParameter1_twoParams", "observableParameter2_twoParams"]

    # noise placeholder
    assert petab.get_formula_placeholders(
        "3.0 * noiseParameter1_oneParam", "oneParam", "noise"
    ) == ["noiseParameter1_oneParam"]

    # multiple instances and in 'wrong' order
    assert petab.get_formula_placeholders(
        "observableParameter2_twoParams * "
        "observableParameter1_twoParams + "
        "otherParam / observableParameter2_twoParams",
        "twoParams",
        "observable",
    ) == ["observableParameter1_twoParams", "observableParameter2_twoParams"]

    # non-consecutive numbering
    with pytest.raises(AssertionError):
        petab.get_formula_placeholders(
            "observableParameter2_twoParams + observableParameter2_twoParams",
            "twoParams",
            "observable",
        )

    # empty
    assert petab.get_formula_placeholders("", "any", "observable") == []

    # non-string
    assert petab.get_formula_placeholders(1, "any", "observable") == []


def test_create_observable_df():
    """Test observables.create_measurement_df."""
    df = petab.create_observable_df()
    assert set(df.columns.values) == set(OBSERVABLE_DF_COLS)


def test_get_placeholders():
    """Test get_placeholders"""
    observable_df = pd.DataFrame(
        data={
            OBSERVABLE_ID: ["obs_1", "obs_2"],
            OBSERVABLE_FORMULA: [
                "observableParameter1_obs_1 * 2 * foo",
                "1 + observableParameter1_obs_2",
            ],
        }
    ).set_index(OBSERVABLE_ID)

    # test with missing noiseFormula
    expected = ["observableParameter1_obs_1", "observableParameter1_obs_2"]
    actual = petab.get_placeholders(observable_df)
    assert actual == expected

    # test with noiseFormula
    observable_df[NOISE_FORMULA] = ["noiseParameter1_obs_1", "2.0"]
    expected = [
        "observableParameter1_obs_1",
        "noiseParameter1_obs_1",
        "observableParameter1_obs_2",
    ]
    actual = petab.get_placeholders(observable_df)
    assert actual == expected
